# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

# To be removed once Investing.com aligns financial products information
OUTDATED2UPDATED = {
    "Prev. Close": "prevClose",
    "Day's Range": "dailyRange",
    "ROI (TTM)": "roiTtm",
    "Revenue": "revenue",
    "Open": "open",
    "Price": "price",
    "Price Open": "priceOpen",
    "Price Range": "priceRange",
    "Coupon": "coupon",
    "Coupon Frequency": "couponFreq",
    "Next Coupon Date": "nextCouponDate",
    "52 wk Range": "weekRange",
    "EPS": "eps",
    "Volume": "volume",
    "Market Cap": "marketCap",
    "Dividend (Yield)": "dividend",
    "Dividend Yield": "dividend",
    "Average Vol. (3m)": "avgVolume",
    "P/E Ratio": "ratio",
    "Beta": "beta",
    "Month": "month",
    "Duration": "duration",
    "1-Year Change": "oneYearReturn",
    "Maturity Date": "maturityDate",
    "Shares Outstanding": "sharesOutstanding",
    "Next Earnings Date": "nextEarningDate",
    "Dividends (TTM)": "dividednsTtm",
    "Total Assets": "totalAssets",
    "Asset Class": "assetClass",
    "Tick Size": "tickSize",
    "Contract Size": "contractSize",
    "Tick Value": "tickValue",
    "Settlement": "settlement",
    "Base Symbol": "baseSymbol",
    "Settlement Day": "settlementDay",
    "Settlement Type": "settlementType",
    "Point Value": "pointValue",
    "Last Rollover Day": "lastRolloverDay",
    "Months": "months",
    "Bid": "bid",
    "Ask": "ask",
    "Rating": "rating",
    "Risk Rating": "riskRating",
    "Issuer": "issuer",
    "Inception Date": "inceptionDate",
    "Min. Investment": "minInvestment",
    "TTM Yield": "ttmYield",
    "Yield": "yield",
    "Turnover": "turnover",
    "ROE": "roe",
    "ROA": "roa",
    "Type": "type",
    "Expenses": "expenses",
    "Category": "category",
    "Leverage": "leverage",
    "Strike Price": "strikePrice",
    "Issue Amount": "issueAmount",
    "Issue Date": "issueDate",
}

FINANCIAL_SUMMARY_TYPES = {
    "income_statement": 0,
    "balance_sheet": 1,
    "cash_flow_statement": 2,
}

FINANCIAL_SUMMARY_PERIODS = {"annual": "Annual", "quarterly": "Interim"}

PRODUCT_FILTERS = {
    "indices": "indice",
    "stocks": "equities",
    "etfs": "etf",
    "funds": "fund",
    "commodities": "commodity",
    "currencies": "currency",
    "cryptos": "crypto",
    "bonds": "bond",
    "certificates": "certificate",
    "fxfutures": "fxfuture",
}

PAIR_FILTERS = {
    "indice": "indices",
    "equities": "stocks",
    "etf": "etfs",
    "fund": "funds",
    "commodity": "commodities",
    "currency": "currencies",
    "crypto": "cryptos",
    "bond": "bonds",
    "certificate": "certificates",
    "fxfuture": "fxfutures",
}

FLAG_FILTERS = {
    "Andorra": "andorra",
    "Argentina": "argentina",
    "Australia": "australia",
    "Austria": "austria",
    "Bahrain": "bahrain",
    "Bangladesh": "bangladesh",
    "Belgium": "belgium",
    "Bermuda": "bermuda",
    "Bosnia": "bosnia",
    "Botswana": "botswana",
    "Brazil": "brazil",
    "Bulgaria": "bulgaria",
    "Canada": "canada",
    "Cayman_Islands": "cayman islands",
    "Chile": "chile",
    "China": "china",
    "Colombia": "colombia",
    "Costa_Rica": "costa rica",
    "Cote_dIvoire": "ivory coast",
    "Croatia": "croatia",
    "Cyprus": "cyprus",
    "Czech_Republic": "czech republic",
    "Denmark": "denmark",
    "Ecuador": "ecuador",
    "Egypt": "egypt",
    "Estonia": "estonia",
    "Europe": "euro zone",
    "Finland": "finland",
    "France": "france",
    "Germany": "germany",
    "Gibraltar": "gibraltar",
    "Greece": "greece",
    "Hong_Kong": "hong kong",
    "Hungary": "hungary",
    "Iceland": "iceland",
    "India": "india",
    "Indonesia": "indonesia",
    "Iraq": "iraq",
    "Ireland": "ireland",
    "Israel": "israel",
    "Italy": "italy",
    "Jamaica": "jamaica",
    "Japan": "japan",
    "Jordan": "jordan",
    "Kazakhstan": "kazakhstan",
    "Kenya": "kenya",
    "Kuwait": "kuwait",
    "Latvia": "latvia",
    "Lebanon": "lebanon",
    "Liechtenstein": "liechtenstein",
    "Lithuania": "lithuania",
    "Luxembourg": "luxembourg",
    "Malawi": "malawi",
    "Malaysia": "malaysia",
    "Malta": "malta",
    "Mauritius": "mauritius",
    "Mexico": "mexico",
    "Monaco": "monaco",
    "Mongolia": "mongolia",
    "Montenegro": "montenegro",
    "Morocco": "morocco",
    "Namibia": "namibia",
    "Netherlands": "netherlands",
    "New_Zealand": "new zealand",
    "Nigeria": "nigeria",
    "Norway": "norway",
    "Oman": "oman",
    "Pakistan": "pakistan",
    "Palestine": "palestine",
    "Peru": "peru",
    "Philippines": "philippines",
    "Poland": "poland",
    "Portugal": "portugal",
    "Qatar": "qatar",
    "Romania": "romania",
    "Russian_Federation": "russia",
    "Rwanda": "rwanda",
    "Saudi_Arabia": "saudi arabia",
    "Serbia": "serbia",
    "Singapore": "singapore",
    "Slovakia": "slovakia",
    "Slovenia": "slovenia",
    "South_Africa": "south africa",
    "South_Korea": "south korea",
    "Spain": "spain",
    "Sri_Lanka": "sri lanka",
    "Sweden": "sweden",
    "Switzerland": "switzerland",
    "Taiwan": "taiwan",
    "Tanzania": "tanzania",
    "Thailand": "thailand",
    "Tunisia": "tunisia",
    "Turkey": "turkey",
    "Uganda": "uganda",
    "Ukraine": "ukraine",
    "Dubai": "dubai",
    "UK": "united kingdom",
    "USA": "united states",
    "Venezuela": "venezuela",
    "Vietnam": "vietnam",
    "Zambia": "zambia",
    "Zimbabwe": "zimbabwe",
}

COUNTRY_FILTERS = {
    "andorra": "Andorra",
    "argentina": "Argentina",
    "australia": "Australia",
    "austria": "Austria",
    "bahrain": "Bahrain",
    "bangladesh": "Bangladesh",
    "belgium": "Belgium",
    "bermuda": "Bermuda",
    "bosnia": "Bosnia",
    "botswana": "Botswana",
    "brazil": "Brazil",
    "bulgaria": "Bulgaria",
    "canada": "Canada",
    "cayman islands": "Cayman_Islands",
    "chile": "Chile",
    "china": "China",
    "colombia": "Colombia",
    "costa rica": "Costa_Rica",
    "ivory coast": "Cote_dIvoire",
    "croatia": "Croatia",
    "cyprus": "Cyprus",
    "czech republic": "Czech_Republic",
    "denmark": "Denmark",
    "ecuador": "Ecuador",
    "egypt": "Egypt",
    "estonia": "Estonia",
    "euro zone": "Europe",
    "finland": "Finland",
    "france": "France",
    "germany": "Germany",
    "gibraltar": "Gibraltar",
    "greece": "Greece",
    "hong kong": "Hong_Kong",
    "hungary": "Hungary",
    "iceland": "Iceland",
    "india": "India",
    "indonesia": "Indonesia",
    "iraq": "Iraq",
    "ireland": "Ireland",
    "israel": "Israel",
    "italy": "Italy",
    "jamaica": "Jamaica",
    "japan": "Japan",
    "jordan": "Jordan",
    "kazakhstan": "Kazakhstan",
    "kenya": "Kenya",
    "kuwait": "Kuwait",
    "latvia": "Latvia",
    "lebanon": "Lebanon",
    "liechtenstein": "Liechtenstein",
    "lithuania": "Lithuania",
    "luxembourg": "Luxembourg",
    "malawi": "Malawi",
    "malaysia": "Malaysia",
    "malta": "Malta",
    "mauritius": "Mauritius",
    "mexico": "Mexico",
    "monaco": "Monaco",
    "mongolia": "Mongolia",
    "montenegro": "Montenegro",
    "morocco": "Morocco",
    "namibia": "Namibia",
    "netherlands": "Netherlands",
    "new zealand": "New_Zealand",
    "nigeria": "Nigeria",
    "norway": "Norway",
    "oman": "Oman",
    "pakistan": "Pakistan",
    "palestine": "Palestine",
    "peru": "Peru",
    "philippines": "Philippines",
    "poland": "Poland",
    "portugal": "Portugal",
    "qatar": "Qatar",
    "romania": "Romania",
    "russia": "Russian_Federation",
    "rwanda": "Rwanda",
    "saudi arabia": "Saudi_Arabia",
    "serbia": "Serbia",
    "singapore": "Singapore",
    "slovakia": "Slovakia",
    "slovenia": "Slovenia",
    "south africa": "South_Africa",
    "south korea": "South_Korea",
    "spain": "Spain",
    "sri lanka": "Sri_Lanka",
    "sweden": "Sweden",
    "switzerland": "Switzerland",
    "taiwan": "Taiwan",
    "tanzania": "Tanzania",
    "thailand": "Thailand",
    "tunisia": "Tunisia",
    "turkey": "Turkey",
    "uganda": "Uganda",
    "ukraine": "Ukraine",
    "dubai": "Dubai",
    "united kingdom": "UK",
    "united states": "USA",
    "venezuela": "Venezuela",
    "vietnam": "Vietnam",
    "zambia": "Zambia",
    "zimbabwe": "Zimbabwe",
}

CATEGORY_FILTERS = {
    "credit": "_credit",
    "employment": "_employment",
    "economic_activity": "_economicActivity",
    "inflation": "_inflation",
    "central_banks": "_centralBanks",
    "confidence": "_confidenceIndex",
    "balance": "_balance",
    "bonds": "_Bonds",
}

IMPORTANCE_RATINGS = {1: "low", 2: "medium", 3: "high"}

COUNTRY_ID_FILTERS = {
    "argentina": 29,
    "australia": 25,
    "austria": 54,
    "bahrain": 145,
    "bangladesh": 47,
    "belgium": 34,
    "bosnia": 174,
    "botswana": 163,
    "brazil": 32,
    "bulgaria": 70,
    "canada": 6,
    "cayman islands": 232,
    "chile": 27,
    "china": 37,
    "colombia": 122,
    "costa rica": 15,
    "croatia": 113,
    "cyprus": 107,
    "czech republic": 55,
    "denmark": 24,
    "dubai": 143,
    "ecuador": 121,
    "egypt": 59,
    "estonia": 89,
    "euro zone": 72,
    "finland": 71,
    "france": 22,
    "germany": 17,
    "greece": 51,
    "hong kong": 39,
    "hungary": 93,
    "iceland": 106,
    "india": 14,
    "indonesia": 48,
    "iraq": 66,
    "ireland": 33,
    "israel": 23,
    "italy": 10,
    "ivory coast": 78,
    "jamaica": 119,
    "japan": 35,
    "jordan": 92,
    "kazakhstan": 102,
    "kenya": 57,
    "kuwait": 94,
    "latvia": 97,
    "lebanon": 68,
    "lithuania": 96,
    "luxembourg": 103,
    "malawi": 111,
    "malaysia": 42,
    "malta": 109,
    "mauritius": 188,
    "mexico": 7,
    "mongolia": 139,
    "montenegro": 247,
    "morocco": 105,
    "namibia": 172,
    "netherlands": 21,
    "new zealand": 43,
    "nigeria": 20,
    "norway": 60,
    "oman": 87,
    "pakistan": 44,
    "palestine": 193,
    "peru": 125,
    "philippines": 45,
    "poland": 53,
    "portugal": 38,
    "qatar": 170,
    "romania": 100,
    "russia": 56,
    "rwanda": 80,
    "saudi arabia": 52,
    "serbia": 238,
    "singapore": 36,
    "slovakia": 90,
    "slovenia": 112,
    "south africa": 110,
    "south korea": 11,
    "spain": 26,
    "sri lanka": 162,
    "sweden": 9,
    "switzerland": 12,
    "taiwan": 46,
    "tanzania": 85,
    "thailand": 41,
    "tunisia": 202,
    "turkey": 63,
    "uganda": 123,
    "ukraine": 61,
    "united kingdom": 4,
    "united states": 5,
    "venezuela": 138,
    "vietnam": 178,
    "zambia": 84,
    "zimbabwe": 75,
}

TIME_FILTERS = {"time_remaining": "timeRemain", "time_only": "timeOnly"}

TIMEZONES = {
    "GMT -11:00": [2, 35],
    "GMT -10:00": [3],
    "GMT -9:00": [4],
    "GMT -8:00": [36, 5],
    "GMT -7:00": [37, 38, 6],
    "GMT -6:00": [39, 7, 40, 41],
    "GMT -5:00": [42, 8, 43],
    "GMT -4:00": [10, 9, 45, 46],
    "GMT -3:30": [11],
    "GMT -3:00": [44, 12, 48, 49, 50, 51, 47],
    "GMT -1:00": [14, 53],
    "GMT": [15, 56],
    "GMT +1:00": [16, 57, 58, 54, 166, 59, 60],
    "GMT +2:00": [62, 64, 65, 66, 68, 17, 67, 61],
    "GMT +3:00": [71, 63, 70, 18, 72],
    "GMT +3:30": [19],
    "GMT +4:00": [20, 73],
    "GMT +4:30": [21],
    "GMT +5:00": [22, 77],
    "GMT +5:30": [23, 79],
    "GMT +5:45": [24],
    "GMT +6:00": [25],
    "GMT +6:30": [26],
    "GMT +7:00": [27],
    "GMT +8:00": [178, 28, 113],
    "GMT +9:00": [29, 88],
    "GMT +9:30": [90],
    "GMT +10:30": [30],
    "GMT +11:00": [31, 32],
    "GMT +12:00": [1],
    "GMT +13:00": [33],
}

PRODUCT_TYPE_FILES = {
    "certificate": "certificates.csv",
    "commodity": "commodities.csv",
    "currency_cross": "currency_crosses.csv",
    "etf": "etfs.csv",
    "fund": "funds.csv",
    "index": "indices.csv",
    "stock": "stocks.csv",
    "bond": "bonds.csv",
}

INTERVAL_FILTERS = {
    "1min": 60,
    "5mins": 60 * 5,
    "15mins": 60 * 15,
    "30mins": 60 * 30,
    "1hour": 60 * 60,
    "5hours": 60 * 60 * 5,
    "daily": 60 * 60 * 24,
    "weekly": "week",
    "monthly": "month",
}

# https://www.investing.com/funds/vanguard-500-index-admiral-technical
FUNDS_INTERVAL_FILTERS = {"daily": 60 * 60 * 24, "weekly": "week", "monthly": "month"}

USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1b3) Gecko/20090305"
    " Firefox/3.1b3 GTB5",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; ko; rv:1.9.1b2) Gecko/20081201"
    " Firefox/3.1b2",
    "Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.9b5) Gecko/2008032620 Firefox/3.0b5",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.8.1.12) Gecko/20080214"
    " Firefox/2.0.0.12",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; cs; rv:1.9.0.8) Gecko/2009032609"
    " Firefox/3.0.8",
    "Mozilla/5.0 (X11; U; OpenBSD i386; en-US; rv:1.8.0.5) Gecko/20060819"
    " Firefox/1.5.0.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.0; es-ES; rv:1.8.0.3) Gecko/20060426"
    " Firefox/1.5.0.3",
    "Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.7.9) Gecko/20050711 Firefox/1.0.6",
    "Mozilla/5.0 (Windows; Windows NT 6.1; rv:2.0b2) Gecko/20100720 Firefox/4.0b2",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b4) Gecko/20100818 Firefox/4.0b4",
    "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2) Gecko/20100308 Ubuntu/10.04"
    " (lucid) Firefox/3.6 GTB7.1",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b7) Gecko/20101111 Firefox/4.0b7",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b8pre) Gecko/20101114 Firefox/4.0b8pre",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b9pre) Gecko/20110111 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.0b9pre) Gecko/20101228"
    " Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.2a1pre) Gecko/20110324"
    " Firefox/4.2a1pre",
    "Mozilla/5.0 (X11; U; Linux amd64; rv:5.0) Gecko/20100101 Firefox/5.0 (Debian)",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:6.0a2) Gecko/20110613 Firefox/6.0a2",
    "Mozilla/5.0 (X11; Linux i686 on x86_64; rv:12.0) Gecko/20100101 Firefox/12.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:15.0) Gecko/20120716 Firefox/15.0a2",
    "Mozilla/5.0 (X11; Ubuntu; Linux armv7l; rv:17.0) Gecko/20100101 Firefox/17.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130328 Firefox/21.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:22.0) Gecko/20130328 Firefox/22.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Mozilla/5.0 (X11; Linux i686; rv:30.0) Gecko/20100101 Firefox/30.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:31.0) Gecko/20100101 Firefox/31.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:58.0) Gecko/20100101"
    " Firefox/58.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like"
    " Gecko) Chrome/1.0.154.53 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like"
    " Gecko) Chrome/1.0.154.36 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.10 (KHTML, like"
    " Gecko) Chrome/7.0.540.0 Safari/534.10",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.4 (KHTML, like"
    " Gecko) Chrome/6.0.481.0 Safari/534.4",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US) AppleWebKit/533.4 (KHTML, like"
    " Gecko) Chrome/5.0.375.86 Safari/533.4",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.2 (KHTML, like"
    " Gecko) Chrome/4.0.223.3 Safari/532.2",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/532.0 (KHTML, like"
    " Gecko) Chrome/4.0.201.1 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/532.0 (KHTML, like"
    " Gecko) Chrome/3.0.195.27 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/530.5 (KHTML, like"
    " Gecko) Chrome/2.0.173.1 Safari/530.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.10 (KHTML, like"
    " Gecko) Chrome/8.0.558.0 Safari/534.10",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/540.0 (KHTML,like Gecko)"
    " Chrome/9.1.0.0 Safari/540.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/534.14 (KHTML, like"
    " Gecko) Chrome/9.0.600.0 Safari/534.14",
    "Mozilla/5.0 (X11; U; Windows NT 6; en-US) AppleWebKit/534.12 (KHTML, like Gecko)"
    " Chrome/9.0.587.0 Safari/534.12",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.13 (KHTML, like"
    " Gecko) Chrome/9.0.597.0 Safari/534.13",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.16 (KHTML, like"
    " Gecko) Chrome/10.0.648.11 Safari/534.16",
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/534.20 (KHTML, like"
    " Gecko) Chrome/11.0.672.2 Safari/534.20",
    "Mozilla/5.0 (Windows NT 6.0) AppleWebKit/535.1 (KHTML, like Gecko)"
    " Chrome/14.0.792.0 Safari/535.1",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/535.2 (KHTML, like Gecko)"
    " Chrome/15.0.872.0 Safari/535.2",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.7 (KHTML, like Gecko)"
    " Chrome/16.0.912.36 Safari/535.7",
    "Mozilla/5.0 (Windows NT 6.0; WOW64) AppleWebKit/535.11 (KHTML, like Gecko)"
    " Chrome/17.0.963.66 Safari/535.11",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.19 (KHTML, like"
    " Gecko) Chrome/18.0.1025.45 Safari/535.19",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/535.24 (KHTML, like Gecko)"
    " Chrome/19.0.1055.1 Safari/535.24",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.6 (KHTML, like Gecko)"
    " Chrome/20.0.1090.0 Safari/536.6",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko)"
    " Chrome/22.0.1207.1 Safari/537.1",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.15 (KHTML, like Gecko)"
    " Chrome/24.0.1295.0 Safari/537.15",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/27.0.1453.93 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/28.0.1467.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/30.0.1599.101 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/31.0.1623.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/34.0.1847.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/37.0.2062.103 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like"
    " Gecko) Chrome/40.0.2214.38 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/46.0.2490.71 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/51.0.2704.103 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/62.0.3202.62 Safari/537.36",
]

BOND_COUNTRIES = [
    {"exchange_id": 37, "tag": "argentina", "country_id": 29, "country": "argentina"},
    {"exchange_id": 18, "tag": "australia", "country_id": 25, "country": "australia"},
    {"exchange_id": 17, "tag": "austria", "country_id": 54, "country": "austria"},
    {"exchange_id": 39, "tag": "bahrain", "country_id": 145, "country": "bahrain"},
    {
        "exchange_id": 115,
        "tag": "bangladesh",
        "country_id": 47,
        "country": "bangladesh",
    },
    {"exchange_id": 14, "tag": "belgium", "country_id": 34, "country": "belgium"},
    {"exchange_id": 77, "tag": "botswana", "country_id": 163, "country": "botswana"},
    {"exchange_id": 47, "tag": "brazil", "country_id": 32, "country": "brazil"},
    {"exchange_id": 44, "tag": "bulgaria", "country_id": 70, "country": "bulgaria"},
    {"exchange_id": 51, "tag": "canada", "country_id": 6, "country": "canada"},
    {"exchange_id": 42, "tag": "chile", "country_id": 27, "country": "chile"},
    {"exchange_id": 54, "tag": "china", "country_id": 37, "country": "china"},
    {"exchange_id": 55, "tag": "colombia", "country_id": 122, "country": "colombia"},
    {"exchange_id": 45, "tag": "croatia", "country_id": 113, "country": "croatia"},
    {"exchange_id": 78, "tag": "cyprus", "country_id": 107, "country": "cyprus"},
    {
        "exchange_id": 56,
        "tag": "czech-republic",
        "country_id": 55,
        "country": "czech republic",
    },
    {"exchange_id": 15, "tag": "denmark", "country_id": 24, "country": "denmark"},
    {"exchange_id": 33, "tag": "egypt", "country_id": 59, "country": "egypt"},
    {"exchange_id": 16, "tag": "finland", "country_id": 71, "country": "finland"},
    {"exchange_id": 9, "tag": "france", "country_id": 22, "country": "france"},
    {"exchange_id": 4, "tag": "germany", "country_id": 17, "country": "germany"},
    {"exchange_id": 41, "tag": "greece", "country_id": 51, "country": "greece"},
    {"exchange_id": 21, "tag": "hong-kong", "country_id": 39, "country": "hong kong"},
    {"exchange_id": 23, "tag": "hungary", "country_id": 93, "country": "hungary"},
    {"exchange_id": 52, "tag": "iceland", "country_id": 106, "country": "iceland"},
    {"exchange_id": 46, "tag": "india", "country_id": 14, "country": "india"},
    {"exchange_id": 57, "tag": "indonesia", "country_id": 48, "country": "indonesia"},
    {"exchange_id": 58, "tag": "ireland", "country_id": 33, "country": "ireland"},
    {"exchange_id": 26, "tag": "israel", "country_id": 23, "country": "israel"},
    {"exchange_id": 6, "tag": "italy", "country_id": 10, "country": "italy"},
    {"exchange_id": 20, "tag": "japan", "country_id": 35, "country": "japan"},
    {"exchange_id": 29, "tag": "jordan", "country_id": 92, "country": "jordan"},
    {"exchange_id": 59, "tag": "kenya", "country_id": 57, "country": "kenya"},
    {"exchange_id": 80, "tag": "latvia", "country_id": 97, "country": "latvia"},
    {"exchange_id": 81, "tag": "lithuania", "country_id": 96, "country": "lithuania"},
    {
        "exchange_id": 61,
        "tag": "luxembourg",
        "country_id": 103,
        "country": "luxembourg",
    },
    {"exchange_id": 62, "tag": "malaysia", "country_id": 42, "country": "malaysia"},
    {"exchange_id": 82, "tag": "malta", "country_id": 109, "country": "malta"},
    {"exchange_id": 86, "tag": "mauritius", "country_id": 188, "country": "mauritius"},
    {"exchange_id": 53, "tag": "mexico", "country_id": 7, "country": "mexico"},
    {"exchange_id": 36, "tag": "morocco", "country_id": 105, "country": "morocco"},
    {"exchange_id": 75, "tag": "namibia", "country_id": 172, "country": "namibia"},
    {
        "exchange_id": 7,
        "tag": "netherlands",
        "country_id": 21,
        "country": "netherlands",
    },
    {
        "exchange_id": 83,
        "tag": "new-zealand",
        "country_id": 43,
        "country": "new zealand",
    },
    {"exchange_id": 96, "tag": "nigeria", "country_id": 20, "country": "nigeria"},
    {"exchange_id": 8, "tag": "norway", "country_id": 60, "country": "norway"},
    {"exchange_id": 63, "tag": "pakistan", "country_id": 44, "country": "pakistan"},
    {"exchange_id": 64, "tag": "peru", "country_id": 125, "country": "peru"},
    {
        "exchange_id": 65,
        "tag": "philippines",
        "country_id": 45,
        "country": "philippines",
    },
    {"exchange_id": 25, "tag": "poland", "country_id": 53, "country": "poland"},
    {"exchange_id": 10, "tag": "portugal", "country_id": 38, "country": "portugal"},
    {"exchange_id": 35, "tag": "qatar", "country_id": 170, "country": "qatar"},
    {"exchange_id": 43, "tag": "romania", "country_id": 100, "country": "romania"},
    {"exchange_id": 40, "tag": "russia", "country_id": 56, "country": "russia"},
    {
        "exchange_id": 28,
        "tag": "saudi-arabia",
        "country_id": 52,
        "country": "saudi arabia",
    },
    {"exchange_id": 102, "tag": "serbia", "country_id": 238, "country": "serbia"},
    {"exchange_id": 19, "tag": "singapore", "country_id": 36, "country": "singapore"},
    {"exchange_id": 84, "tag": "slovakia", "country_id": 90, "country": "slovakia"},
    {"exchange_id": 66, "tag": "slovenia", "country_id": 112, "country": "slovenia"},
    {
        "exchange_id": 22,
        "tag": "south-africa",
        "country_id": 110,
        "country": "south africa",
    },
    {
        "exchange_id": 60,
        "tag": "south-korea",
        "country_id": 11,
        "country": "south korea",
    },
    {"exchange_id": 11, "tag": "spain", "country_id": 26, "country": "spain"},
    {"exchange_id": 67, "tag": "sri-lanka", "country_id": 162, "country": "sri lanka"},
    {"exchange_id": 12, "tag": "sweden", "country_id": 9, "country": "sweden"},
    {
        "exchange_id": 5,
        "tag": "switzerland",
        "country_id": 12,
        "country": "switzerland",
    },
    {"exchange_id": 68, "tag": "taiwan", "country_id": 46, "country": "taiwan"},
    {"exchange_id": 69, "tag": "thailand", "country_id": 41, "country": "thailand"},
    {"exchange_id": 49, "tag": "turkey", "country_id": 63, "country": "turkey"},
    {"exchange_id": 76, "tag": "uganda", "country_id": 123, "country": "uganda"},
    {"exchange_id": 70, "tag": "ukraine", "country_id": 61, "country": "ukraine"},
    {"exchange_id": 3, "tag": "uk", "country_id": 4, "country": "united kingdom"},
    {"exchange_id": 1, "tag": "usa", "country_id": 5, "country": "united states"},
    {"exchange_id": 71, "tag": "venezuela", "country_id": 138, "country": "venezuela"},
    {"exchange_id": 72, "tag": "vietnam", "country_id": 178, "country": "vietnam"},
]

CERTIFICATE_COUNTRIES = [
    {"country": "france", "id": 22},
    {"country": "germany", "id": 17},
    {"country": "italy", "id": 10},
    {"country": "netherlands", "id": 21},
    {"country": "sweden", "id": 9},
]

CURRENCIES = {
    "AED": 22,
    "AFN": 93,
    "ALL": 58,
    "AMD": 158,
    "ANG": 95,
    "AOA": 159,
    "ARS": 32,
    "AUD": 1,
    "AZN": 161,
    "BAM": 96,
    "BBD": 33,
    "BDT": 59,
    "BGN": 91,
    "BHD": 36,
    "BIF": 98,
    "BND": 37,
    "BOB": 99,
    "BRL": 35,
    "BRLT": 1896,
    "BSD": 38,
    "BWP": 39,
    "BYN": 100,
    "BZD": 101,
    "CAD": 15,
    "CHF": 4,
    "CLP": 40,
    "CNH": 194,
    "CNY": 41,
    "COP": 92,
    "CRC": 42,
    "CUP": 102,
    "CVE": 188,
    "CZK": 16,
    "DJF": 44,
    "DKK": 6,
    "DOP": 45,
    "DZD": 104,
    "EGP": 47,
    "ETB": 48,
    "EUR": 17,
    "FJD": 49,
    "GBP": 3,
    "GEL": 183,
    "GHS": 50,
    "GMD": 51,
    "GNF": 107,
    "GTQ": 52,
    "HKD": 20,
    "HNL": 53,
    "HRK": 109,
    "HTG": 54,
    "HUF": 11,
    "IDR": 55,
    "ILS": 23,
    "INR": 29,
    "IQD": 56,
    "IRR": 57,
    "ISK": 60,
    "JMD": 61,
    "JOD": 24,
    "JPY": 2,
    "KES": 110,
    "KGS": 168,
    "KHR": 111,
    "KMF": 169,
    "KRW": 28,
    "KWD": 25,
    "KYD": 62,
    "KZT": 63,
    "LAK": 113,
    "LBP": 114,
    "LKR": 115,
    "LSL": 185,
    "LYD": 119,
    "MAD": 64,
    "MDL": 181,
    "MGA": 121,
    "MKD": 123,
    "MMK": 124,
    "MNT": 196,
    "MOP": 125,
    "MRU": 126,
    "MUR": 66,
    "MVR": 170,
    "MWK": 67,
    "MXN": 14,
    "MYR": 68,
    "MZN": 192,
    "NAD": 128,
    "NGN": 69,
    "NIO": 70,
    "NOK": 7,
    "NPR": 129,
    "NZD": 5,
    "OMR": 71,
    "PAB": 130,
    "PEN": 72,
    "PGK": 73,
    "PHP": 74,
    "PKR": 75,
    "PLN": 8,
    "PYG": 131,
    "QAR": 76,
    "RON": 78,
    "RSD": 133,
    "RUB": 79,
    "RWF": 134,
    "SAR": 26,
    "SCR": 81,
    "SDG": 136,
    "SEK": 18,
    "SGD": 19,
    "SOS": 139,
    "STN": 141,
    "SVC": 142,
    "SYP": 143,
    "SZL": 82,
    "THB": 10,
    "TJS": 193,
    "TMT": 1746,
    "TND": 83,
    "TRY": 9,
    "TTD": 84,
    "TWD": 34,
    "TZS": 145,
    "UAH": 85,
    "UGX": 146,
    "USD": 12,
    "UYU": 147,
    "UZS": 148,
    "VES": 149,
    "VND": 87,
    "VUV": 177,
    "XAF": 150,
    "XAGg": 216,
    "XBR": 1683,
    "XCD": 179,
    "XOF": 151,
    "XPF": 88,
    "YER": 153,
    "ZAR": 13,
    "ZMW": 155,
}

ETF_COUNTRIES = [
    {"country": "australia", "code": "au"},
    {"country": "austria", "code": "at"},
    {"country": "belgium", "code": "be"},
    {"country": "brazil", "code": "br"},
    {"country": "bulgaria", "code": "bg"},
    {"country": "canada", "code": "ca"},
    {"country": "china", "code": "cn"},
    {"country": "colombia", "code": "co"},
    {"country": "denmark", "code": "dk"},
    {"country": "egypt", "code": "eg"},
    {"country": "euro zone", "code": "eu"},
    {"country": "finland", "code": "fi"},
    {"country": "france", "code": "fr"},
    {"country": "germany", "code": "de"},
    {"country": "greece", "code": "gr"},
    {"country": "hong kong", "code": "hk"},
    {"country": "hungary", "code": "hu"},
    {"country": "iceland", "code": "is"},
    {"country": "india", "code": "in"},
    {"country": "indonesia", "code": "id"},
    {"country": "ireland", "code": "ie"},
    {"country": "israel", "code": "il"},
    {"country": "italy", "code": "it"},
    {"country": "japan", "code": "jp"},
    {"country": "malaysia", "code": "my"},
    {"country": "mexico", "code": "mx"},
    {"country": "netherlands", "code": "nl"},
    {"country": "new zealand", "code": "nz"},
    {"country": "nigeria", "code": "ng"},
    {"country": "norway", "code": "no"},
    {"country": "peru", "code": "pe"},
    {"country": "poland", "code": "pl"},
    {"country": "portugal", "code": "pt"},
    {"country": "qatar", "code": "qa"},
    {"country": "romania", "code": "ro"},
    {"country": "russia", "code": "ru"},
    {"country": "saudi arabia", "code": "sa"},
    {"country": "singapore", "code": "sg"},
    {"country": "south africa", "code": "sa"},
    {"country": "south korea", "code": "kr"},
    {"country": "spain", "code": "es"},
    {"country": "sweden", "code": "se"},
    {"country": "switzerland", "code": "ch"},
    {"country": "taiwan", "code": "tw"},
    {"country": "thailand", "code": "th"},
    {"country": "turkey", "code": "tr"},
    {"country": "united kingdom", "code": "uk"},
    {"country": "united states", "code": "us"},
    {"country": "vietnam", "code": "vn"},
    {"country": "euro zone", "code": ""},
]

FUND_COUNTRIES = [
    {"country": "andorra", "id": 206},
    {"country": "australia", "id": 25},
    {"country": "austria", "id": 54},
    {"country": "bahrain", "id": 145},
    {"country": "belgium", "id": 34},
    {"country": "bermuda", "id": 8},
    {"country": "brazil", "id": 32},
    {"country": "canada", "id": 6},
    {"country": "cayman islands", "id": 232},
    {"country": "chile", "id": 27},
    {"country": "china", "id": 37},
    {"country": "czech republic", "id": 55},
    {"country": "denmark", "id": 24},
    {"country": "estonia", "id": 89},
    {"country": "finland", "id": 71},
    {"country": "france", "id": 22},
    {"country": "germany", "id": 17},
    {"country": "gibraltar", "id": 147},
    {"country": "greece", "id": 51},
    {"country": "hong kong", "id": 39},
    {"country": "hungary", "id": 93},
    {"country": "iceland", "id": 106},
    {"country": "india", "id": 14},
    {"country": "indonesia", "id": 48},
    {"country": "ireland", "id": 33},
    {"country": "israel", "id": 23},
    {"country": "italy", "id": 10},
    {"country": "japan", "id": 35},
    {"country": "latvia", "id": 97},
    {"country": "liechtenstein", "id": 190},
    {"country": "lithuania", "id": 96},
    {"country": "luxembourg", "id": 103},
    {"country": "malaysia", "id": 42},
    {"country": "malta", "id": 109},
    {"country": "mauritius", "id": 188},
    {"country": "mexico", "id": 7},
    {"country": "monaco", "id": 115},
    {"country": "namibia", "id": 172},
    {"country": "netherlands", "id": 21},
    {"country": "new zealand", "id": 43},
    {"country": "norway", "id": 60},
    {"country": "oman", "id": 87},
    {"country": "pakistan", "id": 44},
    {"country": "philippines", "id": 45},
    {"country": "poland", "id": 53},
    {"country": "portugal", "id": 38},
    {"country": "qatar", "id": 170},
    {"country": "russia", "id": 56},
    {"country": "saudi arabia", "id": 52},
    {"country": "singapore", "id": 36},
    {"country": "slovenia", "id": 112},
    {"country": "south africa", "id": 110},
    {"country": "south korea", "id": 11},
    {"country": "spain", "id": 26},
    {"country": "sweden", "id": 9},
    {"country": "switzerland", "id": 12},
    {"country": "taiwan", "id": 46},
    {"country": "thailand", "id": 41},
    {"country": "dubai", "id": 143},
    {"country": "united kingdom", "id": 4},
    {"country": "united states", "id": 5},
    {"country": "vietnam", "id": 178},
]

STOCK_COUNTRIES = {
    "argentina": 10141,
    "brazil": 602,
    "canada": 608,
    "chile": 10150,
    "colombia": 723,
    "costa rica": 688,
    "jamaica": 10210,
    "mexico": 612,
    "peru": 722,
    "united states": 800,
    "venezuela": 25920001,
    "austria": 10124,
    "belgium": 10121,
    "bosnia": 804,
    "bulgaria": 10151,
    "croatia": 10148,
    "cyprus": 725,
    "czech republic": 614,
    "denmark": 10122,
    "finland": 10123,
    "france": 10112,
    "germany": 10106,
    "greece": 10145,
    "hungary": 10128,
    "iceland": 610,
    "ireland": 616,
    "italy": 10109,
    "luxembourg": 802,
    "malta": 803,
    "montenegro": 10213,
    "netherlands": 10110,
    "norway": 10111,
    "poland": 10130,
    "portugal": 10113,
    "romania": 10149,
    "russia": 10144,
    "serbia": 805,
    "slovakia": 625,
    "slovenia": 620,
    "spain": 10119,
    "sweden": 10120,
    "switzerland": 10107,
    "turkey": 606,
    "ukraine": 623,
    "united kingdom": 801,
    "australia": 10125,
    "bangladesh": 10201,
    "china": 724,
    "hong kong": 10126,
    "india": 600,
    "indonesia": 615,
    "japan": 692,
    "kazakhstan": 10207,
    "malaysia": 618,
    "mongolia": 10216,
    "new zealand": 27525828,
    "pakistan": 619,
    "philippines": 721,
    "singapore": 10129,
    "south korea": 694,
    "sri lanka": 621,
    "taiwan": 622,
    "thailand": 720,
    "vietnam": 624,
    "bahrain": 10143,
    "egypt": 10136,
    "iraq": 10204,
    "israel": 10131,
    "jordan": 10134,
    "kuwait": 10132,
    "lebanon": 10146,
    "oman": 10142,
    "palestine": 10137,
    "qatar": 10139,
    "saudi arabia": 10133,
    "dubai": 10135,
    "botswana": 656,
    "ivory coast": 729,
    "kenya": 617,
    "malawi": 680,
    "mauritius": 678,
    "morocco": 10147,
    "namibia": 652,
    "nigeria": 628,
    "rwanda": 686,
    "south africa": 10127,
    "tanzania": 682,
    "tunisia": 604,
    "uganda": 654,
    "zambia": 25957877,
    "zimbabwe": 684,
}

INDEX_COUNTRIES = [
    {"country": "argentina", "country_name": "argentina"},
    {"country": "australia", "country_name": "australia"},
    {"country": "austria", "country_name": "austria"},
    {"country": "bahrain", "country_name": "bahrain"},
    {"country": "bangladesh", "country_name": "bangladesh"},
    {"country": "belgium", "country_name": "belgium"},
    {"country": "bosnia", "country_name": "bosnia"},
    {"country": "botswana", "country_name": "botswana"},
    {"country": "brazil", "country_name": "brazil"},
    {"country": "bulgaria", "country_name": "bulgaria"},
    {"country": "canada", "country_name": "canada"},
    {"country": "chile", "country_name": "chile"},
    {"country": "china", "country_name": "china"},
    {"country": "colombia", "country_name": "colombia"},
    {"country": "costa rica", "country_name": "costa rica"},
    {"country": "ivory coast", "country_name": "ivory coast"},
    {"country": "croatia", "country_name": "croatia"},
    {"country": "cyprus", "country_name": "cyprus"},
    {"country": "czech republic", "country_name": "czech republic"},
    {"country": "denmark", "country_name": "denmark"},
    {"country": "ecuador", "country_name": "ecuador"},
    {"country": "egypt", "country_name": "egypt"},
    {"country": "estonia", "country_name": "estonia"},
    {"country": "finland", "country_name": "finland"},
    {"country": "france", "country_name": "france"},
    {"country": "germany", "country_name": "germany"},
    {"country": "greece", "country_name": "greece"},
    {"country": "hong kong", "country_name": "hong kong"},
    {"country": "hungary", "country_name": "hungary"},
    {"country": "iceland", "country_name": "iceland"},
    {"country": "india", "country_name": "india"},
    {"country": "indonesia", "country_name": "indonesia"},
    {"country": "iraq", "country_name": "iraq"},
    {"country": "ireland", "country_name": "ireland"},
    {"country": "israeli", "country_name": "israel"},
    {"country": "italy", "country_name": "italy"},
    {"country": "jamaica", "country_name": "jamaica"},
    {"country": "japan", "country_name": "japan"},
    {"country": "jordan", "country_name": "jordan"},
    {"country": "kazakhstan", "country_name": "kazakhstan"},
    {"country": "kenya", "country_name": "kenya"},
    {"country": "kuwaiti", "country_name": "kuwait"},
    {"country": "latvia", "country_name": "latvia"},
    {"country": "lebanon", "country_name": "lebanon"},
    {"country": "lithuania", "country_name": "lithuania"},
    {"country": "luxembourg", "country_name": "luxembourg"},
    {"country": "malawi", "country_name": "malawi"},
    {"country": "malaysia", "country_name": "malaysia"},
    {"country": "malta", "country_name": "malta"},
    {"country": "mauritius", "country_name": "mauritius"},
    {"country": "mexico", "country_name": "mexico"},
    {"country": "mongolia", "country_name": "mongolia"},
    {"country": "montenegro", "country_name": "montenegro"},
    {"country": "morocco", "country_name": "morocco"},
    {"country": "namibia", "country_name": "namibia"},
    {"country": "netherlands", "country_name": "netherlands"},
    {"country": "new zealand", "country_name": "new zealand"},
    {"country": "nigeria", "country_name": "nigeria"},
    {"country": "norway", "country_name": "norway"},
    {"country": "oman", "country_name": "oman"},
    {"country": "pakistan", "country_name": "pakistan"},
    {"country": "palestine", "country_name": "palestinian territory"},
    {"country": "peru", "country_name": "peru"},
    {"country": "philippines", "country_name": "philippines"},
    {"country": "poland", "country_name": "poland"},
    {"country": "portugal", "country_name": "portugal"},
    {"country": "qatar", "country_name": "qatar"},
    {"country": "romania", "country_name": "romania"},
    {"country": "russia", "country_name": "russia"},
    {"country": "rwanda", "country_name": "rwanda"},
    {"country": "saudi arabia", "country_name": "saudi arabia"},
    {"country": "serbia", "country_name": "serbia"},
    {"country": "singapore", "country_name": "singapore"},
    {"country": "slovakia", "country_name": "slovakia"},
    {"country": "slovenia", "country_name": "slovenia"},
    {"country": "south africa", "country_name": "south africa"},
    {"country": "south korea", "country_name": "south korea"},
    {"country": "spain", "country_name": "spain"},
    {"country": "sri lanka", "country_name": "sri lanka"},
    {"country": "sweden", "country_name": "sweden"},
    {"country": "switzerland", "country_name": "switzerland"},
    {"country": "taiwan", "country_name": "taiwan"},
    {"country": "tanzania", "country_name": "tanzania"},
    {"country": "thailand", "country_name": "thailand"},
    {"country": "tunisia", "country_name": "tunisia"},
    {"country": "turkey", "country_name": "turkey"},
    {"country": "uganda", "country_name": "uganda"},
    {"country": "ukraine", "country_name": "ukraine"},
    {"country": "dubai", "country_name": "dubai"},
    {"country": "uk", "country_name": "united kingdom"},
    {"country": "usa", "country_name": "united states"},
    {"country": "venezuela", "country_name": "venezuela"},
    {"country": "vietnam", "country_name": "vietnam"},
    {"country": "world", "country_name": "world"},
    {"country": "zambia", "country_name": "zambia"},
    {"country": "zimbabwe", "country_name": "zimbabwe"},
    {"country": "euro zone", "country_name": "euro zone"},
]



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json
from datetime import date, datetime, timedelta
from random import randint

import pandas as pd
import pkg_resources
import pytz
import requests
from lxml.html import fromstring
from unidecode import unidecode

from .data.indices_data import (
    index_countries_as_list,
    indices_as_df,
    indices_as_dict,
    indices_as_list,
)
from .utils.data import Data
from .utils.extra import random_user_agent


def get_indices(country=None):
    """
    This function retrieves all the available `indices` from Investing.com as previously listed in investpy, and
    returns them as a :obj:`pandas.DataFrame` with all the information of every available index. If the country
    filtering is applied, just the indices from the introduced country are going to be returned.
    All the available indices can be found at: https://www.investing.com/indices/world-indices and at
    https://www.investing.com/indices/world-indices, since both world and global indices are retrieved.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available indices from.

    Returns:
        :obj:`pandas.DataFrame` - indices_df:
            The resulting :obj:`pandas.DataFrame` contains all the indices information retrieved from Investing.com,
            as previously listed by investpy.

            In case the information was successfully retrieved, the :obj:`pandas.DataFrame` will look like::

                country | name | full_name | symbol | currency | class | market
                --------|------|-----------|--------|----------|-------|--------
                xxxxxxx | xxxx | xxxxxxxxx | xxxxxx | xxxxxxxx | xxxxx | xxxxxx

    Raises:
        ValueError: raised if any of the introduced parameters is missing or errored.
        FileNotFoundError: raised if the `indices.csv` file was not found.
        IOError: raised if the `indices.csv` file from `investpy` is missing or errored.

    """

    return indices_as_df(country=country)


def get_indices_list(country=None):
    """
    This function retrieves all the available `indices` from Investing.com as previously listed in investpy, and
    returns them as a :obj:`list` with the names of every available index. If the country filtering is applied, just
    the indices from the introduced country are going to be returned.
    All the available indices can be found at: https://www.investing.com/indices/world-indices and at
    https://www.investing.com/indices/world-indices, since both world and global indices are retrieved.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available indices from.

    Returns:
        :obj:`list` - indices_list:
            The resulting :obj:`list` contains the retrieved data, which corresponds to the index names of
            every index listed in Investing.com.

            In case the information was successfully retrieved, the :obj:`list` will look like::

                indices = ['S&P Merval', 'S&P Merval Argentina', 'S&P/BYMA Argentina General', ...]

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if the `indices.csv` file was not found.
        IOError: raised if the `indices.csv` file is missing or errored.

    """

    return indices_as_list(country=country)


def get_indices_dict(country=None, columns=None, as_json=False):
    """
    This function retrieves all the available `indices` from Investing.com as previously listed in investpy, and
    returns them as a :obj:`dict` with all the information of every available index. If the country
    filtering is applied, just the indices from the introduced country are going to be returned. Additionally, the
    columns to retrieve data from can be specified as a parameter formatted as a :obj:`list`.
    All the available indices can be found at: https://www.investing.com/indices/world-indices and at
    https://www.investing.com/indices/world-indices, since both world and global indices are retrieved.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available indices from.
        columns (:obj:`list` of :obj:`str`, optional): description
            a :obj:`list` containing the column names from which the data is going to be retrieved.
        as_json (:obj:`bool`, optional): description
            value to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`dict` or :obj:`json` - indices_dict:
            The resulting :obj:`dict` contains the retrieved data if found, if not, the corresponding
            fields are filled with `None` values.

            In case the information was successfully retrieved, the :obj:`dict` will look like::

                indices_dict = {
                    'country': country,
                    'name': name,
                    'full_name': full_name,
                    'symbol': symbol,
                    'currency': currency,
                    'class': class,
                    'market': market
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if the `indices.csv` file was not found.
        IOError: raised if the `indices.csv` file is missing or errored.

    """

    return indices_as_dict(country=country, columns=columns, as_json=as_json)


def get_index_countries():
    """
    This function retrieves all the country names indexed in Investing.com with available indices to retrieve data
    from, via reading the `indices.csv` file from the resources directory. So on, this function will display a listing
    containing a set of countries, in order to let the user know which countries are available for indices data retrieval.

    Returns:
        :obj:`list` - countries:
            The resulting :obj:`list` contains all the available countries with indices as indexed in Investing.com

    Raises:
        FileNotFoundError: raised if the `indices.csv` file was not found.
        IOError: raised if the `indices.csv` file is missing or errored.

    """

    return index_countries_as_list()


def get_index_recent_data(
    index, country, as_json=False, order="ascending", interval="Daily"
):
    """
    This function retrieves recent historical data from the introduced `index` from Investing
    via Web Scraping. The resulting data can it either be stored in a :obj:`pandas.DataFrame` or in a
    :obj:`json` file, with `ascending` or `descending` order.

    Args:
        index (:obj:`str`): name of the index to retrieve recent historical data from.
        country (:obj:`str`): name of the country from where the index is.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        order (:obj:`str`, optional):
            optional argument to define the order of the retrieved data (`ascending`, `asc` or `descending`, `desc`).
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function returns either a :obj:`pandas.DataFrame` or a :obj:`json` file containing the retrieved
            recent data from the specified index via argument. The dataset contains the open, high, low, close and volume
            values for the selected index on market days, additionally the currency value is returned.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but if we define `as_json=True`, then the output will be::

                {
                    name: name,
                    recent: [
                        {
                            date: dd/mm/yyyy,
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised if there was an argument error.
        IOError: raised if indices object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced index does not match any of the indexed ones.
        ConnectionError: raised if GET requests does not return 200 status code.
        IndexError: raised if index information was unavailable or not found.

    Examples:
        >>> data = investpy.get_index_recent_data(index='ibex 35', country='spain')
        >>> data.head()
                       Open     High      Low    Close   Volume Currency
        Date
        2019-08-26  12604.7  12646.3  12510.4  12621.3  4770000      EUR
        2019-08-27  12618.3  12723.3  12593.6  12683.8  8230000      EUR
        2019-08-28  12657.2  12697.2  12585.1  12642.5  7300000      EUR
        2019-08-29  12637.2  12806.6  12633.8  12806.6  5650000      EUR
        2019-08-30  12767.6  12905.9  12756.9  12821.6  6040000      EUR

    """

    if not index:
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if not isinstance(index, str):
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "indices.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        indices = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0059: indices file not found or errored.")

    if indices is None:
        raise IOError("ERR#0037: indices not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_index_countries():
        raise RuntimeError(
            "ERR#0034: country " + country + " not found, check if it is correct."
        )

    indices = indices[indices["country"] == country]

    index = unidecode(index.strip().lower())

    if index not in list(indices["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0045: index " + index + " not found, check if it is correct."
        )

    full_name = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "full_name"
    ]
    id_ = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "id"
    ]
    name = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "name"
    ]

    index_currency = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "currency"
    ]

    header = full_name + " Historical Data"

    params = {
        "curr_id": id_,
        "smlID": str(randint(1000000, 99999999)),
        "header": header,
        "interval_sec": interval.capitalize(),
        "sort_col": "date",
        "sort_ord": "DESC",
        "action": "historical_data",
    }

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/instruments/HistoricalDataAjax"

    req = requests.post(url, headers=head, data=params)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

    result = list()

    if path_:
        for elements_ in path_:
            if elements_.xpath(".//td")[0].text_content() == "No results found":
                raise IndexError(
                    "ERR#0046: index information unavailable or not found."
                )

            info = []

            for nested_ in elements_.xpath(".//td"):
                info.append(nested_.get("data-real-value"))

            index_date = datetime.strptime(
                str(
                    datetime.fromtimestamp(int(info[0]), tz=pytz.timezone("GMT")).date()
                ),
                "%Y-%m-%d",
            )

            index_close = float(info[1].replace(",", ""))
            index_open = float(info[2].replace(",", ""))
            index_high = float(info[3].replace(",", ""))
            index_low = float(info[4].replace(",", ""))

            index_volume = int(info[5])

            result.insert(
                len(result),
                Data(
                    index_date,
                    index_open,
                    index_high,
                    index_low,
                    index_close,
                    index_volume,
                    index_currency,
                    None,
                ),
            )

        if order in ["ascending", "asc"]:
            result = result[::-1]
        elif order in ["descending", "desc"]:
            result = result

        if as_json is True:
            json_ = {
                "name": name,
                "recent": [value.index_as_json() for value in result],
            }

            return json.dumps(json_, sort_keys=False)
        elif as_json is False:
            df = pd.DataFrame.from_records([value.index_to_dict() for value in result])
            df.set_index("Date", inplace=True)

            return df
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_index_historical_data(
    index,
    country,
    from_date,
    to_date,
    as_json=False,
    order="ascending",
    interval="Daily",
):
    """
    This function retrieves historical data of the introduced `index` (from the specified country, note that both
    index and country should match since if the introduced index is not listed in the indices of that country, the
    function will raise an error). The retrieved historical data are the OHLC values plus the Volume and the Currency in
    which those values are specified, from the introduced date range if valid. So on, the resulting data can it either be
    stored in a :obj:`pandas.DataFrame` or in a :obj:`json` file.

    Args:
        index (:obj:`str`): name of the index to retrieve recent historical data from.
        country (:obj:`str`): name of the country from where the index is.
        from_date (:obj:`str`): date as `str` formatted as `dd/mm/yyyy`, from where data is going to be retrieved.
        to_date (:obj:`str`): date as `str` formatted as `dd/mm/yyyy`, until where data is going to be retrieved.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function returns either a :obj:`pandas.DataFrame` or a :obj:`json` file containing the retrieved
            historical data from the specified index via argument. The dataset contains the open, high, low, close and
            volume values for the selected index on market days, additionally the currency in which those values are
            specified is returned.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but if we define `as_json=True`, then the output will be::

                {
                    name: name,
                    historical: [
                        {
                            date: dd/mm/yyyy,
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised if there was an argument error.
        IOError: raised if indices object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced index does not match any of the indexed ones.
        ConnectionError: raised if GET requests does not return 200 status code.
        IndexError: raised if index information was unavailable or not found.

    Examples:
        >>> data = investpy.get_index_historical_data(index='ibex 35', country='spain', from_date='01/01/2018', to_date='01/01/2019')
        >>> data.head()
                       Open     High      Low    Close    Volume Currency
        Date
        2018-01-02  15128.2  15136.7  14996.6  15096.8  10340000      EUR
        2018-01-03  15145.0  15186.9  15091.9  15106.9  12800000      EUR
        2018-01-04  15105.5  15368.7  15103.7  15368.7  17070000      EUR
        2018-01-05  15353.9  15407.5  15348.6  15398.9  11180000      EUR
        2018-01-08  15437.1  15448.7  15344.0  15373.3  12890000      EUR

    """

    if not index:
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if not isinstance(index, str):
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    try:
        datetime.strptime(from_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError("ERR#0011: incorrect data format, it should be 'dd/mm/yyyy'.")

    try:
        datetime.strptime(to_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError("ERR#0011: incorrect data format, it should be 'dd/mm/yyyy'.")

    start_date = datetime.strptime(from_date, "%d/%m/%Y")
    end_date = datetime.strptime(to_date, "%d/%m/%Y")

    if start_date >= end_date:
        raise ValueError(
            "ERR#0032: to_date should be greater than from_date, both formatted as"
            " 'dd/mm/yyyy'."
        )

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    date_interval = {
        "intervals": [],
    }

    flag = True

    while flag is True:
        diff = end_date.year - start_date.year

        if diff > 19:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": start_date.replace(year=start_date.year + 19).strftime(
                    "%m/%d/%Y"
                ),
            }

            date_interval["intervals"].append(obj)

            start_date = start_date.replace(year=start_date.year + 19) + timedelta(
                days=1
            )
        else:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": end_date.strftime("%m/%d/%Y"),
            }

            date_interval["intervals"].append(obj)

            flag = False

    interval_limit = len(date_interval["intervals"])
    interval_counter = 0

    data_flag = False

    resource_package = "investpy"
    resource_path = "/".join(("resources", "indices.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        indices = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0059: indices file not found or errored.")

    if indices is None:
        raise IOError("ERR#0037: indices not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_index_countries():
        raise RuntimeError(
            "ERR#0034: country " + country + " not found, check if it is correct."
        )

    indices = indices[indices["country"] == country]

    index = unidecode(index.strip().lower())

    if index not in list(indices["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0045: index " + index + " not found, check if it is correct."
        )

    full_name = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "full_name"
    ]
    id_ = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "id"
    ]
    name = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "name"
    ]

    index_currency = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "currency"
    ]

    final = list()

    header = full_name + " Historical Data"

    for index in range(len(date_interval["intervals"])):
        interval_counter += 1

        params = {
            "curr_id": id_,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "st_date": date_interval["intervals"][index]["start"],
            "end_date": date_interval["intervals"][index]["end"],
            "interval_sec": interval.capitalize(),
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        head = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        url = "https://www.investing.com/instruments/HistoricalDataAjax"

        req = requests.post(url, headers=head, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                "ERR#0015: error " + str(req.status_code) + ", try again later."
            )

        if not req.text:
            continue

        root_ = fromstring(req.text)
        path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

        result = list()

        if path_:
            for elements_ in path_:
                if elements_.xpath(".//td")[0].text_content() == "No results found":
                    if interval_counter < interval_limit:
                        data_flag = False
                    else:
                        raise IndexError(
                            "ERR#0046: index information unavailable or not found."
                        )
                else:
                    data_flag = True

                info = []

                for nested_ in elements_.xpath(".//td"):
                    info.append(nested_.get("data-real-value"))

                if data_flag is True:
                    index_date = datetime.strptime(
                        str(
                            datetime.fromtimestamp(
                                int(info[0]), tz=pytz.timezone("GMT")
                            ).date()
                        ),
                        "%Y-%m-%d",
                    )

                    index_close = float(info[1].replace(",", ""))
                    index_open = float(info[2].replace(",", ""))
                    index_high = float(info[3].replace(",", ""))
                    index_low = float(info[4].replace(",", ""))

                    index_volume = int(info[5])

                    result.insert(
                        len(result),
                        Data(
                            index_date,
                            index_open,
                            index_high,
                            index_low,
                            index_close,
                            index_volume,
                            index_currency,
                            None,
                        ),
                    )
            if data_flag is True:
                if order in ["ascending", "asc"]:
                    result = result[::-1]
                elif order in ["descending", "desc"]:
                    result = result

                if as_json is True:
                    json_list = [value.index_as_json() for value in result]

                    final.append(json_list)
                elif as_json is False:
                    df = pd.DataFrame.from_records(
                        [value.index_to_dict() for value in result]
                    )
                    df.set_index("Date", inplace=True)

                    final.append(df)
        else:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    if order in ["descending", "desc"]:
        final.reverse()

    if as_json is True:
        json_ = {
            "name": name,
            "historical": [value for json_list in final for value in json_list],
        }
        return json.dumps(json_, sort_keys=False)
    elif as_json is False:
        return pd.concat(final)


def get_index_information(index, country, as_json=False):
    """
    This function retrieves fundamental financial information from the specified index. The retrieved
    information from the index can be valuable as it is additional information that can be used combined
    with OHLC values, so to determine financial insights from the company which holds the specified index.

    Args:
        index (:obj:`str`): name of the index to retrieve recent historical data from.
        country (:obj:`str`): name of the country from where the index is.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`pandas.DataFrame` or :obj:`dict`- index_information:
            The resulting :obj:`pandas.DataFrame` contains the information fields retrieved from Investing.com
            from the specified index; it can also be returned as a :obj:`dict`, if argument `as_json=True`.

            If any of the information fields could not be retrieved, that field/s will be filled with
            None values. If the retrieval process succeeded, the resulting :obj:`dict` will look like::

                index_information = {
                    "Index Name": "S&P Merval",
                    "Prev. Close": 36769.59,
                    "Volume": None,
                    "Todays Range": "36,769.59 - 37,894.32",
                    "Open": 36769.59,
                    "Average Vol. (3m)": None,
                    "52 wk Range": "22,484.4 - 44,470.76",
                    "1-Year Change": "18.19%"
                }

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `indices.csv` file was not found or errored.
        IOError: raised if `indices.csv` file is empty or errored.
        RuntimeError: raised if scraping process failed while running.
        ConnectionError: raised if the connection to Investing.com errored (did not return HTTP 200)

    """

    if not index:
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if not isinstance(index, str):
        raise ValueError("ERR#0047: index param is mandatory and should be a str.")

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "indices.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        indices = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0059: indices file not found or errored.")

    if indices is None:
        raise IOError("ERR#0037: indices not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_index_countries():
        raise RuntimeError(
            "ERR#0034: country " + country + " not found, check if it is correct."
        )

    indices = indices[indices["country"] == country]

    index = unidecode(index.strip().lower())

    if index not in list(indices["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0045: index " + index + " not found, check if it is correct."
        )

    name = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "name"
    ]
    tag = indices.loc[
        (indices["name"].apply(unidecode).str.lower() == index).idxmax(), "tag"
    ]

    url = "https://www.investing.com/indices/" + tag

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath("//dl[@data-test='key-info']/div")

    result = pd.DataFrame(
        columns=[
            "Index Name",
            "Prev. Close",
            "Volume",
            "Day's Range",
            "Open",
            "Average Vol. (3m)",
            "52 wk Range",
            "1-Year Change",
        ]
    )
    result.at[0, "Index Name"] = name

    if not path_:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    for elements_ in path_:
        title_ = elements_.xpath(".//dt")[0].text_content()
        element = elements_.xpath(".//dd")[0]
        if title_ in result.columns.tolist():
            try:
                result.at[0, title_] = float(element.text_content().replace(",", ""))
                continue
            except:
                pass
            try:
                text = element.text_content().strip()
                result.at[0, title_] = datetime.strptime(text, "%b %d, %Y").strftime(
                    "%d/%m/%Y"
                )
                continue
            except:
                pass
            try:
                value = element.text_content().strip()
                if value.__contains__("K"):
                    value = float(value.replace("K", "").replace(",", "")) * 1e3
                elif value.__contains__("M"):
                    value = float(value.replace("M", "").replace(",", "")) * 1e6
                elif value.__contains__("B"):
                    value = float(value.replace("B", "").replace(",", "")) * 1e9
                elif value.__contains__("T"):
                    value = float(value.replace("T", "").replace(",", "")) * 1e12
                result.at[0, title_] = value
                continue
            except:
                pass

    result.replace({"N/A": None}, inplace=True)

    if as_json is True:
        json_ = result.iloc[0].to_dict()
        return json_
    elif as_json is False:
        return result


def get_indices_overview(country, as_json=False, n_results=100):
    """
    This function retrieves an overview containing all the real time data available for the main indices from a country,
    such as the names, symbols, current value, etc. as indexed in Investing.com. So on, the main usage of this
    function is to get an overview on the main indices from a country, so to get a general view. Note that since
    this function is retrieving a lot of information at once, by default just the overview of the Top 100 indices
    is being retrieved, but an additional parameter called n_results can be specified so to retrieve N results.

    Args:
        country (:obj:`str`): name of the country to retrieve the indices overview from.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        n_results (:obj:`int`, optional): number of results to be displayed on the overview table (0-1000).

    Returns:
        :obj:`pandas.DataFrame` - indices_overview:
            The resulting :obj:`pandas.DataFrame` contains all the data available in Investing.com of the main indices
            from a country in order to get an overview of it.

            If the retrieval process succeeded, the resulting :obj:`pandas.DataFrame` should look like::

                country | name | last | high | low | change | change_percentage | currency
                --------|------|------|------|-----|--------|-------------------|----------
                xxxxxxx | xxxx | xxxx | xxxx | xxx | xxxxxx | xxxxxxxxxxxxxxxxx | xxxxxxxx

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised when `indices.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError:
            raised either if the introduced country does not match any of the listed ones or if no overview results could be
            retrieved from Investing.com.
        ConnectionError: raised if GET requests does not return 200 status code.

    """

    if country is None:
        raise ValueError("ERR#0039: country can not be None, it should be a str.")

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if not isinstance(n_results, int):
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    if 1 > n_results or n_results > 1000:
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "indices.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        indices = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0059: indices file not found or errored.")

    if indices is None:
        raise IOError("ERR#0037: indices not found or unable to retrieve.")

    country = unidecode(country.strip().lower())

    if country not in get_index_countries():
        raise ValueError(
            "ERR#0034: country " + country + " not found, check if it is correct."
        )

    indices = indices[indices["country"] == country]

    if country == "united states":
        country = "usa"
    elif country == "united kingdom":
        country = "uk"

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = (
        "https://www.investing.com/indices/"
        + country.replace(" ", "-")
        + "-indices?&majorIndices=on&primarySectors=on&additionalIndices=on&otherIndices=on"
    )

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    table = root_.xpath(".//table[@id='cr1']/tbody/tr")

    results = list()

    if len(table) > 0:
        for row in table[:n_results]:
            id_ = row.get("id").replace("pair_", "")
            country_check = (
                row.xpath(".//td[@class='flag']/span")[0].get("title").lower()
            )

            if country_check == "bosnia-herzegovina":
                country_check = "bosnia"
            elif country_check == "palestinian territory":
                country_check = "palestine"
            elif country_check == "united arab emirates":
                country_check = "dubai"
            elif country_check == "cote d'ivoire":
                country_check = "ivory coast"

            name = (
                row.xpath(".//td[contains(@class, 'elp')]/a")[0].text_content().strip()
            )

            pid = "pid-" + id_

            last = row.xpath(".//td[@class='" + pid + "-last']")[0].text_content()
            high = row.xpath(".//td[@class='" + pid + "-high']")[0].text_content()
            low = row.xpath(".//td[@class='" + pid + "-low']")[0].text_content()

            pc = row.xpath(".//td[contains(@class, '" + pid + "-pc')]")[
                0
            ].text_content()
            pcp = row.xpath(".//td[contains(@class, '" + pid + "-pcp')]")[
                0
            ].text_content()

            data = {
                "country": country_check,
                "name": name,
                "last": float(last.replace(",", "")),
                "high": float(high.replace(",", "")),
                "low": float(low.replace(",", "")),
                "change": pc,
                "change_percentage": pcp,
                "currency": indices.loc[(indices["name"] == name).idxmax(), "currency"],
            }

            results.append(data)
    else:
        raise RuntimeError(
            "ERR#0092: no data found while retrieving the overview from Investing.com"
        )

    df = pd.DataFrame(results)

    if as_json:
        return json.loads(df.to_json(orient="records"))
    else:
        return df


def search_indices(by, value):
    """
    This function searches indices by the introduced value for the specified field. This means that this function
    is going to search if there is a value that matches the introduced value for the specified field which is the
    `indices.csv` column name to search in. Available fields to search indices are 'name', 'full_name' and 'symbol'.

    Args:
       by (:obj:`str`): name of the field to search for, which is the column name ('name', 'full_name' or 'symbol').
       value (:obj:`str`): value of the field to search for, which is the str that is going to be searched.

    Returns:
       :obj:`pandas.DataFrame` - search_result:
           The resulting `pandas.DataFrame` contains the search results from the given query (the specified value
           in the specified field). If there are no results and error will be raised, but otherwise this
           `pandas.DataFrame` will contain all the available field values that match the introduced query.

    Raises:
       ValueError: raised if any of the introduced params is not valid or errored.
       FileNotFoundError: raised if `indices.csv` file is missing.
       IOError: raised if data could not be retrieved due to file error.
       RuntimeError: raised if no results were found for the introduced value in the introduced field.

    """

    if not by:
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not isinstance(by, str):
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not value:
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    if not isinstance(value, str):
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "indices.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        indices = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0059: indices file not found or errored.")

    if indices is None:
        raise IOError("ERR#0037: indices not found or unable to retrieve.")

    indices.drop(columns=["tag", "id"], inplace=True)

    available_search_fields = indices.columns.tolist()

    if isinstance(by, str) and by not in available_search_fields:
        raise ValueError(
            "ERR#0026: the introduced field to search can either just be "
            + " or ".join(available_search_fields)
        )

    indices["matches"] = indices[by].str.contains(value, case=False)

    search_result = indices.loc[indices["matches"] == True].copy()

    if len(search_result) == 0:
        raise RuntimeError(
            "ERR#0043: no results were found for the introduced " + str(by) + "."
        )

    search_result.drop(columns=["matches"], inplace=True)
    search_result.reset_index(drop=True, inplace=True)

    return search_result



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

__author__ = "Alvaro Bartolome @ alvarobartt in GitHub"
__version__ = "1.0.8"

from .bonds import (
    get_bond_countries,
    get_bond_historical_data,
    get_bond_information,
    get_bond_recent_data,
    get_bonds,
    get_bonds_dict,
    get_bonds_list,
    get_bonds_overview,
    search_bonds,
)
from .certificates import (
    get_certificate_countries,
    get_certificate_historical_data,
    get_certificate_information,
    get_certificate_recent_data,
    get_certificates,
    get_certificates_dict,
    get_certificates_list,
    get_certificates_overview,
    search_certificates,
)
from .commodities import (
    get_commodities,
    get_commodities_dict,
    get_commodities_list,
    get_commodities_overview,
    get_commodity_groups,
    get_commodity_historical_data,
    get_commodity_information,
    get_commodity_recent_data,
    search_commodities,
)
from .crypto import (
    get_crypto_historical_data,
    get_crypto_information,
    get_crypto_recent_data,
    get_cryptos,
    get_cryptos_dict,
    get_cryptos_list,
    get_cryptos_overview,
    search_cryptos,
)
from .currency_crosses import (
    get_available_currencies,
    get_currency_cross_historical_data,
    get_currency_cross_information,
    get_currency_cross_recent_data,
    get_currency_crosses,
    get_currency_crosses_dict,
    get_currency_crosses_list,
    get_currency_crosses_overview,
    search_currency_crosses,
)
from .etfs import (
    get_etf_countries,
    get_etf_historical_data,
    get_etf_information,
    get_etf_recent_data,
    get_etfs,
    get_etfs_dict,
    get_etfs_list,
    get_etfs_overview,
    search_etfs,
)
from .funds import (
    get_fund_countries,
    get_fund_historical_data,
    get_fund_information,
    get_fund_recent_data,
    get_funds,
    get_funds_dict,
    get_funds_list,
    get_funds_overview,
    search_funds,
)
from .indices import (
    get_index_countries,
    get_index_historical_data,
    get_index_information,
    get_index_recent_data,
    get_indices,
    get_indices_dict,
    get_indices_list,
    get_indices_overview,
    search_indices,
)
from .news import economic_calendar
from .search import search_quotes
from .stocks import (
    get_stock_company_profile,
    get_stock_countries,
    get_stock_dividends,
    get_stock_financial_summary,
    get_stock_historical_data,
    get_stock_information,
    get_stock_recent_data,
    get_stocks,
    get_stocks_dict,
    get_stocks_list,
    get_stocks_overview,
    search_stocks,
)
from .technical import moving_averages, pivot_points, technical_indicators

# from .search import search_events



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json

import pandas as pd
import pkg_resources
from unidecode import unidecode

from ..utils import constant as cst


def stocks_as_df(country=None):
    """
    This function retrieves all the stock data stored in `stocks.csv` file, which previously was
    retrieved from Investing.com. Since the resulting object is a matrix of data, the stock data is properly
    structured in rows and columns, where columns are the stock data attribute names. Additionally, country
    filtering can be specified, which will make this function return not all the stored stock data, but just
    the stock data of the stocks from the introduced country.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available stocks from.

    Returns:
        :obj:`pandas.DataFrame` - stocks_df:
            The resulting :obj:`pandas.DataFrame` contains all the stock data from the introduced country if specified,
            or from every country if None was specified, as indexed in Investing.com from the information previously
            retrieved by investpy and stored on a csv file.

            So on, the resulting :obj:`pandas.DataFrame` will look like::

                country | name | full name | isin | currency | symbol
                --------|------|-----------|------|----------|--------
                xxxxxxx | xxxx | xxxxxxxxx | xxxx | xxxxxxxx | xxxxxx

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `stocks.csv` file was not found.
        IOError: raised when `stocks.csv` file is missing or empty.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    resource_package = "investpy"
    resource_path = "/".join((("resources", "stocks.csv")))
    if pkg_resources.resource_exists(resource_package, resource_path):
        stocks = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0056: stocks file not found or errored.")

    if stocks is None:
        raise IOError("ERR#0001: stocks list not found or unable to retrieve.")

    stocks.drop(columns=["tag", "id"], inplace=True)
    stocks = stocks.where(pd.notnull(stocks), None)

    if country is None:
        stocks.reset_index(drop=True, inplace=True)
        return stocks
    else:
        country = unidecode(country.strip().lower())

        if country not in stock_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        stocks = stocks[stocks["country"] == country]
        stocks.reset_index(drop=True, inplace=True)

        return stocks


def stocks_as_list(country=None):
    """
    This function retrieves all the stock symbols stored in `stocks.csv` file, which contains all the
    data from the stocks as previously retrieved from Investing.com. So on, this function will just return
    the stock symbols which will be one of the input parameters when it comes to stock data retrieval functions
    from investpy. Additionally, note that the country filtering can be applied, which is really useful since
    this function just returns the symbols and in stock data retrieval functions both the symbol and the country
    must be specified and they must match.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available stocks from.

    Returns:
        :obj:`list` - stocks_list:
            The resulting :obj:`list` contains the all the stock symbols from the introduced country if specified,
            or from every country if None was specified, as indexed in Investing.com from the information previously
            retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of stock symbols will look like::

                stocks_list = ['TS', 'APBR', 'GGAL', 'TXAR', 'PAMP', ...]

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `stocks.csv` file was not found.
        IOError: raised when `stocks.csv` file is missing or empty.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    resource_package = "investpy"
    resource_path = "/".join((("resources", "stocks.csv")))
    if pkg_resources.resource_exists(resource_package, resource_path):
        stocks = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0056: stocks file not found or errored.")

    if stocks is None:
        raise IOError("ERR#0001: stocks list not found or unable to retrieve.")

    stocks.drop(columns=["tag", "id"], inplace=True)
    stocks = stocks.where(pd.notnull(stocks), None)

    if country is None:
        return stocks["symbol"].tolist()
    else:
        country = unidecode(country.strip().lower())

        if country not in stock_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        return stocks[stocks["country"] == country]["symbol"].tolist()


def stocks_as_dict(country=None, columns=None, as_json=False):
    """
    This function retrieves all the stock information stored in the `stocks.csv` file and formats it as a
    Python dictionary which contains the same information as the file, but every row is a :obj:`dict` and
    all of them are contained in a :obj:`list`. Note that the dictionary structure is the same one as the
    JSON structure. Some optional paramaters can be specified such as the country, columns or as_json, which
    are a filtering by country so not to return all the stocks but just the ones from the introduced country,
    the column names that want to be retrieved in case of needing just some columns to avoid unnecessary information
    load, and whether the information wants to be returned as a JSON object or as a dictionary; respectively.

    Args:
        country (:obj:`str`, optional): name of the country to retrieve all its available stocks from.
        columns (:obj:`list`, optional):column names of the stock data to retrieve, can be: <country, name, full_name, isin, currency, symbol>
        as_json (:obj:`bool`, optional): if True the returned data will be a :obj:`json` object, if False, a :obj:`list` of :obj:`dict`.

    Returns:
        :obj:`list` of :obj:`dict` OR :obj:`json` - stocks_dict:
            The resulting :obj:`list` of :obj:`dict` contains the retrieved data from every stock as indexed in Investing.com from
            the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of :obj:`dict` will look like::

                stocks_dict = {
                    'country': country,
                    'name': name,
                    'full_name': full_name,
                    'isin': isin,
                    'currency': currency,
                    'symbol': symbol,
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `stocks.csv` file was not found.
        IOError: raised when `stocks.csv` file is missing or empty.

    """

    if country is not None and not isinstance(country, str):
        raise ValueError("ERR#0025: specified country value not valid.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join((("resources", "stocks.csv")))
    if pkg_resources.resource_exists(resource_package, resource_path):
        stocks = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0056: stocks file not found or errored.")

    if stocks is None:
        raise IOError("ERR#0001: stocks list not found or unable to retrieve.")

    stocks.drop(columns=["tag", "id"], inplace=True)
    stocks = stocks.where(pd.notnull(stocks), None)

    if columns is None:
        columns = stocks.columns.tolist()
    else:
        if not isinstance(columns, list):
            raise ValueError(
                "ERR#0020: specified columns argument is not a list, it can just be"
                " list type."
            )

    if not all(column in stocks.columns.tolist() for column in columns):
        raise ValueError(
            "ERR#0021: specified columns does not exist, available columns are "
            "<country, name, full_name, isin, currency, symbol>"
        )

    if country is None:
        if as_json:
            return json.dumps(stocks[columns].to_dict(orient="records"))
        else:
            return stocks[columns].to_dict(orient="records")
    else:
        country = unidecode(country.strip().lower())

        if country not in stock_countries_as_list():
            raise ValueError(
                "ERR#0034: country " + country + " not found, check if it is correct."
            )

        if as_json:
            return json.dumps(
                stocks[stocks["country"] == country][columns].to_dict(orient="records")
            )
        else:
            return stocks[stocks["country"] == country][columns].to_dict(
                orient="records"
            )


def stock_countries_as_list():
    """
    This function returns a listing with all the available countries from where stocks can be retrieved, so to
    let the user know which of them are available, since the parameter country is mandatory in every stock retrieval
    function.

    Returns:
        :obj:`list` - countries:
            The resulting :obj:`list` contains all the available countries with stocks as indexed in Investing.com

    """

    return list(cst.STOCK_COUNTRIES.keys())



# Copyright 2018-2021 Alvaro Bartolome, alvarobartt @ GitHub
# See LICENSE for details.

import json
from datetime import date, datetime, timedelta
from random import randint

import pandas as pd
import pkg_resources
import pytz
import requests
from lxml.html import fromstring
from unidecode import unidecode

from .data.crypto_data import cryptos_as_df, cryptos_as_dict, cryptos_as_list
from .utils.data import Data
from .utils.extra import random_user_agent


def get_cryptos():
    """
    This function retrieves all the crypto data stored in `cryptos.csv` file, which previously was
    retrieved from Investing.com. Since the resulting object is a matrix of data, the crypto data is properly
    structured in rows and columns, where columns are the crypto data attribute names.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Returns:
        :obj:`pandas.DataFrame` - cryptos_df:
            The resulting :obj:`pandas.DataFrame` contains all the crypto data from every available crypto coin as
            indexed in Investing.com from the information previously retrieved by investpy and stored on a csv file.

            So on, the resulting :obj:`pandas.DataFrame` will look like::

                name | symbol | currency
                -----|--------|----------
                xxxx | xxxxxx | xxxxxxxx

    Raises:
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    return cryptos_as_df()


def get_cryptos_list():
    """
    This function retrieves all the crypto coin names stored in `cryptos.csv` file, which contains all the
    data from the crypto coins as previously retrieved from Investing.com. So on, this function will just return
    the crypto coin names which will be the main input parameters when it comes to crypto data retrieval functions
    from investpy.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Returns:
        :obj:`list` - cryptos_list:
            The resulting :obj:`list` contains the all the available crypto coin names as indexed in Investing.com
            from the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of crypto coin names will look like::

                cryptos_list = ['Bitcoin', 'Ethereum', 'XRP', 'Bitcoin Cash', 'Tether', 'Litecoin', ...]

    Raises:
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    return cryptos_as_list()


def get_cryptos_dict(columns=None, as_json=False):
    """
    This function retrieves all the crypto information stored in the `cryptos.csv` file and formats it as a
    Python dictionary which contains the same information as the file, but every row is a :obj:`dict` and
    all of them are contained in a :obj:`list`. Note that the dictionary structure is the same one as the
    JSON structure. Some optional paramaters can be specified such as the columns or as_json, which are the
    column names that want to be retrieved in case of needing just some columns to avoid unnecessary information
    load, and whether the information wants to be returned as a JSON object or as a dictionary; respectively.

    Note that just some cryptos are available for retrieval, since Investing.com does not provide information
    from all the available ones, just the main ones.

    Args:
        columns (:obj:`list`, optional):column names of the crypto data to retrieve, can be: <name, currency, symbol>
        as_json (:obj:`bool`, optional): if True the returned data will be a :obj:`json` object, if False, a :obj:`list` of :obj:`dict`.

    Returns:
        :obj:`list` of :obj:`dict` OR :obj:`json` - cryptos_dict:
            The resulting :obj:`list` of :obj:`dict` contains the retrieved data from every crypto coin as indexed in Investing.com from
            the information previously retrieved by investpy and stored on a csv file.

            In case the information was successfully retrieved, the :obj:`list` of :obj:`dict` will look like::

                cryptos_dict = {
                    'name': name,
                    'currency': currency,
                    'symbol': symbol,
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid.
        FileNotFoundError: raised if `cryptos.csv` file was not found.
        IOError: raised when `cryptos.csv` file is missing or empty.

    """

    return cryptos_as_dict(columns=columns, as_json=as_json)


def get_crypto_recent_data(crypto, as_json=False, order="ascending", interval="Daily"):
    """
    This function retrieves recent historical data from the introduced crypto from Investing.com. So on, the recent data
    of the introduced crypto will be retrieved and returned as a :obj:`pandas.DataFrame` if the parameters are valid
    and the request to Investing.com succeeds. Note that additionally some optional parameters can be specified: as_json
    and order, which let the user decide if the data is going to be returned as a :obj:`json` or not, and if the historical
    data is going to be ordered ascending or descending (where the index is the date), respectively.

    Args:
        crypto (:obj:`str`): name of the crypto currency to retrieve data from.
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function can return either a :obj:`pandas.DataFrame` or a :obj:`json` object, containing the retrieved
            recent data of the specified crypto currency. So on, the resulting dataframe contains the open, high, low,
            close and volume values for the selected crypto on market days and the currency in which those values are presented.

            The resulting recent data, in case that the default parameters were applied, will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but in case that as_json parameter was defined as True, then the output will be::

                {
                    name: name,
                    recent: [
                        {
                            date: 'dd/mm/yyyy',
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        IOError: raised if cryptos object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced crypto name was not found or did not match any of the existing ones.
        ConnectionError: raised if connection to Investing.com could not be established.
        IndexError: raised if crypto recent data was unavailable or not found in Investing.com.

    Examples:
        >>> data = investpy.get_crypto_recent_data(crypto='bitcoin')
        >>> data.head()
                      Open     High     Low   Close   Volume Currency
        Date
        2019-10-25  7422.8   8697.7  7404.9  8658.3  1177632      USD
        2019-10-26  8658.4  10540.0  8061.8  9230.6  1784005      USD
        2019-10-27  9230.6   9773.2  9081.0  9529.6  1155038      USD
        2019-10-28  9530.1   9866.9  9202.5  9207.2  1039295      USD
        2019-10-29  9206.5   9531.3  9125.3  9411.3   918477      USD

    """

    if not crypto:
        raise ValueError(
            "ERR#0083: crypto parameter is mandatory and must be a valid crypto name."
        )

    if not isinstance(crypto, str):
        raise ValueError("ERR#0084: crypto argument needs to be a str.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    crypto = unidecode(crypto.strip().lower())

    if crypto not in list(cryptos["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0085: crypto currency: "
            + crypto
            + ", not found, check if it is correct."
        )

    status = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "status"
    ]
    if status == "unavailable":
        raise ValueError(
            "ERR#0086: the selected crypto currency is not available for retrieval in"
            " Investing.com."
        )

    crypto_name = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "name"
    ]
    crypto_id = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "id"
    ]
    crypto_currency = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "currency"
    ]

    header = crypto_name + " Historical Data"

    params = {
        "curr_id": crypto_id,
        "smlID": str(randint(1000000, 99999999)),
        "header": header,
        "interval_sec": interval.capitalize(),
        "sort_col": "date",
        "sort_ord": "DESC",
        "action": "historical_data",
    }

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/instruments/HistoricalDataAjax"

    req = requests.post(url, headers=head, data=params)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")
    result = list()

    if path_:
        for elements_ in path_:
            if elements_.xpath(".//td")[0].text_content() == "No results found":
                raise IndexError(
                    "ERR#0087: crypto information unavailable or not found."
                )

            info = []

            for nested_ in elements_.xpath(".//td"):
                info.append(nested_.get("data-real-value"))

            crypto_date = datetime.strptime(
                str(
                    datetime.fromtimestamp(int(info[0]), tz=pytz.timezone("GMT")).date()
                ),
                "%Y-%m-%d",
            )

            crypto_close = float(info[1].replace(",", ""))
            crypto_open = float(info[2].replace(",", ""))
            crypto_high = float(info[3].replace(",", ""))
            crypto_low = float(info[4].replace(",", ""))

            crypto_volume = int(info[5])

            result.insert(
                len(result),
                Data(
                    crypto_date,
                    crypto_open,
                    crypto_high,
                    crypto_low,
                    crypto_close,
                    crypto_volume,
                    crypto_currency,
                    None,
                ),
            )

        if order in ["ascending", "asc"]:
            result = result[::-1]
        elif order in ["descending", "desc"]:
            result = result

        if as_json is True:
            json_ = {
                "name": crypto_name,
                "recent": [value.crypto_as_json() for value in result],
            }

            return json.dumps(json_, sort_keys=False)
        elif as_json is False:
            df = pd.DataFrame.from_records([value.crypto_to_dict() for value in result])
            df.set_index("Date", inplace=True)

            return df
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_crypto_historical_data(
    crypto, from_date, to_date, as_json=False, order="ascending", interval="Daily"
):
    """
    This function retrieves historical data from the introduced crypto from Investing.com. So on, the historical data
    of the introduced crypto will be retrieved and returned as a :obj:`pandas.DataFrame` if the parameters are valid
    and the request to Investing.com succeeds. Note that additionally some optional parameters can be specified: as_json
    and order, which let the user decide if the data is going to be returned as a :obj:`json` or not, and if the historical
    data is going to be ordered ascending or descending (where the index is the date), respectively.

    Args:
        crypto (:obj:`str`): name of the crypto currency to retrieve data from.
        from_date (:obj:`str`): date formatted as `dd/mm/yyyy`, since when data is going to be retrieved.
        to_date (:obj:`str`): date formatted as `dd/mm/yyyy`, until when data is going to be retrieved.
        as_json (:obj:`bool`, optional):
            to determine the format of the output data, either a :obj:`pandas.DataFrame` if False and a :obj:`json` if True.
        order (:obj:`str`, optional): to define the order of the retrieved data which can either be ascending or descending.
        interval (:obj:`str`, optional):
            value to define the historical data interval to retrieve, by default `Daily`, but it can also be `Weekly` or `Monthly`.

    Returns:
        :obj:`pandas.DataFrame` or :obj:`json`:
            The function can return either a :obj:`pandas.DataFrame` or a :obj:`json` object, containing the retrieved
            historical data of the specified crypto currency. So on, the resulting dataframe contains the open, high,
            low, close and volume values for the selected crypto on market days and the currency in which those values are presented.

            The returned data is case we use default arguments will look like::

                Date || Open | High | Low | Close | Volume | Currency
                -----||------|------|-----|-------|--------|----------
                xxxx || xxxx | xxxx | xxx | xxxxx | xxxxxx | xxxxxxxx

            but if we define `as_json=True`, then the output will be::

                {
                    name: name,
                    historical: [
                        {
                            date: 'dd/mm/yyyy',
                            open: x,
                            high: x,
                            low: x,
                            close: x,
                            volume: x,
                            currency: x
                        },
                        ...
                    ]
                }

    Raises:
        ValueError: raised whenever any of the introduced arguments is not valid or errored.
        IOError: raised if cryptos object/file was not found or unable to retrieve.
        RuntimeError: raised if the introduced crypto currency name was not found or did not match any of the existing ones.
        ConnectionError: raised if connection to Investing.com could not be established.
        IndexError: raised if crypto historical data was unavailable or not found in Investing.com.

    Examples:
        >>> data = investpy.get_crypto_historical_data(crypto='bitcoin', from_date='01/01/2018', to_date='01/01/2019')
        >>> data.head()
                       Open     High      Low    Close  Volume Currency
        Date
        2018-01-01  13850.5  13921.5  12877.7  13444.9   78425      USD
        2018-01-02  13444.9  15306.1  12934.2  14754.1  137732      USD
        2018-01-03  14754.1  15435.0  14579.7  15156.6  106543      USD
        2018-01-04  15156.5  15408.7  14244.7  15180.1  110969      USD
        2018-01-05  15180.1  17126.9  14832.4  16954.8  141960      USD

    """

    if not crypto:
        raise ValueError(
            "ERR#0083: crypto parameter is mandatory and must be a valid crypto name."
        )

    if not isinstance(crypto, str):
        raise ValueError("ERR#0084: crypto argument needs to be a str.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if order not in ["ascending", "asc", "descending", "desc"]:
        raise ValueError(
            "ERR#0003: order argument can just be ascending (asc) or descending (desc),"
            " str type."
        )

    if not interval:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    if not isinstance(interval, str):
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    interval = interval.lower()

    if interval not in ["daily", "weekly", "monthly"]:
        raise ValueError(
            "ERR#0073: interval value should be a str type and it can just be either"
            " 'Daily', 'Weekly' or 'Monthly'."
        )

    try:
        datetime.strptime(from_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0011: incorrect from_date date format, it should be 'dd/mm/yyyy'."
        )

    try:
        datetime.strptime(to_date, "%d/%m/%Y")
    except ValueError:
        raise ValueError(
            "ERR#0012: incorrect to_date format, it should be 'dd/mm/yyyy'."
        )

    start_date = datetime.strptime(from_date, "%d/%m/%Y")
    end_date = datetime.strptime(to_date, "%d/%m/%Y")

    if start_date >= end_date:
        raise ValueError(
            "ERR#0032: to_date should be greater than from_date, both formatted as"
            " 'dd/mm/yyyy'."
        )

    date_interval = {
        "intervals": [],
    }

    flag = True

    while flag is True:
        diff = end_date.year - start_date.year

        if diff > 19:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": start_date.replace(year=start_date.year + 19).strftime(
                    "%m/%d/%Y"
                ),
            }

            date_interval["intervals"].append(obj)

            start_date = start_date.replace(year=start_date.year + 19) + timedelta(
                days=1
            )
        else:
            obj = {
                "start": start_date.strftime("%m/%d/%Y"),
                "end": end_date.strftime("%m/%d/%Y"),
            }

            date_interval["intervals"].append(obj)

            flag = False

    interval_limit = len(date_interval["intervals"])
    interval_counter = 0

    data_flag = False

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    crypto = unidecode(crypto.strip().lower())

    if crypto not in list(cryptos["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0085: crypto currency: "
            + crypto
            + ", not found, check if it is correct."
        )

    status = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "status"
    ]
    if status == "unavailable":
        raise ValueError(
            "ERR#0086: the selected crypto currency is not available for retrieval in"
            " Investing.com."
        )

    crypto_name = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "name"
    ]
    crypto_id = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "id"
    ]
    crypto_currency = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "currency"
    ]

    header = crypto_name + " Historical Data"

    final = list()

    for index in range(len(date_interval["intervals"])):
        interval_counter += 1

        params = {
            "curr_id": crypto_id,
            "smlID": str(randint(1000000, 99999999)),
            "header": header,
            "st_date": date_interval["intervals"][index]["start"],
            "end_date": date_interval["intervals"][index]["end"],
            "interval_sec": interval.capitalize(),
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        head = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        url = "https://www.investing.com/instruments/HistoricalDataAjax"

        req = requests.post(url, headers=head, data=params)

        if req.status_code != 200:
            raise ConnectionError(
                "ERR#0015: error " + str(req.status_code) + ", try again later."
            )

        if not req.text:
            continue

        root_ = fromstring(req.text)
        path_ = root_.xpath(".//table[@id='curr_table']/tbody/tr")

        result = list()

        if path_:
            for elements_ in path_:
                if elements_.xpath(".//td")[0].text_content() == "No results found":
                    if interval_counter < interval_limit:
                        data_flag = False
                    else:
                        raise IndexError(
                            "ERR#0087: crypto information unavailable or not found."
                        )
                else:
                    data_flag = True

                info = []

                for nested_ in elements_.xpath(".//td"):
                    info.append(nested_.get("data-real-value"))

                if data_flag is True:
                    crypto_date = datetime.strptime(
                        str(
                            datetime.fromtimestamp(
                                int(info[0]), tz=pytz.timezone("GMT")
                            ).date()
                        ),
                        "%Y-%m-%d",
                    )

                    crypto_close = float(info[1].replace(",", ""))
                    crypto_open = float(info[2].replace(",", ""))
                    crypto_high = float(info[3].replace(",", ""))
                    crypto_low = float(info[4].replace(",", ""))

                    crypto_volume = int(info[5] or 0)

                    result.insert(
                        len(result),
                        Data(
                            crypto_date,
                            crypto_open,
                            crypto_high,
                            crypto_low,
                            crypto_close,
                            crypto_volume,
                            crypto_currency,
                            None,
                        ),
                    )

            if data_flag is True:
                if order in ["ascending", "asc"]:
                    result = result[::-1]
                elif order in ["descending", "desc"]:
                    result = result

                if as_json is True:
                    json_list = [value.crypto_as_json() for value in result]

                    final.append(json_list)
                elif as_json is False:
                    df = pd.DataFrame.from_records(
                        [value.crypto_to_dict() for value in result]
                    )
                    df.set_index("Date", inplace=True)

                    final.append(df)
        else:
            raise RuntimeError("ERR#0004: data retrieval error while scraping.")

    if order in ["descending", "desc"]:
        final.reverse()

    if as_json is True:
        json_ = {
            "name": crypto_name,
            "historical": [value for json_list in final for value in json_list],
        }
        return json.dumps(json_, sort_keys=False)
    elif as_json is False:
        return pd.concat(final)


def get_crypto_information(crypto, as_json=False):
    """
    This function retrieves fundamental financial information from the specified crypto currency. The retrieved
    information from the crypto currency can be valuable as it is additional information that can be used combined
    with OHLC values, so to determine financial insights from the company which holds the specified crypto currency.

    Args:
        currency_cross (:obj:`str`): name of the currency_cross to retrieve recent historical data from.
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`dict` or :obj:`json`).

    Returns:
        :obj:`pandas.DataFrame` or :obj:`dict`- crypto_information:
            The resulting :obj:`pandas.DataFrame` contains the information fields retrieved from Investing.com
            from the specified crypto currency; it can also be returned as a :obj:`dict`, if argument `as_json=True`.

            If any of the information fields could not be retrieved, that field/s will be filled with
            None values. If the retrieval process succeeded, the resulting :obj:`dict` will look like::

                crypto_information = {
                    'Chg (7D)': '-4.63%',
                    'Circulating Supply': ' BTC18.10M',
                    'Crypto Currency': 'Bitcoin',
                    'Currency': 'USD',
                    'Market Cap': '$129.01B',
                    'Max Supply': 'BTC21.00M',
                    'Todays Range': '7,057.8 - 7,153.1',
                    'Vol (24H)': '$17.57B'
                }

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        FileNotFoundError: raised if `cryptos.csv` file was not found or errored.
        IOError: raised if `cryptos.csv` file is empty or errored.
        RuntimeError: raised if scraping process failed while running.
        ConnectionError: raised if the connection to Investing.com errored (did not return HTTP 200)

    """

    if not crypto:
        raise ValueError(
            "ERR#0083: crypto parameter is mandatory and must be a valid crypto name."
        )

    if not isinstance(crypto, str):
        raise ValueError("ERR#0084: crypto argument needs to be a str.")

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    crypto = unidecode(crypto.strip().lower())

    if crypto not in list(cryptos["name"].apply(unidecode).str.lower()):
        raise RuntimeError(
            "ERR#0085: crypto currency: "
            + crypto
            + ", not found, check if it is correct."
        )

    status = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "status"
    ]
    if status == "unavailable":
        raise ValueError(
            "ERR#0086: the selected crypto currency is not available for retrieval in"
            " Investing.com."
        )

    name = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "name"
    ]
    currency = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "currency"
    ]
    tag = cryptos.loc[
        (cryptos["name"].apply(unidecode).str.lower() == crypto).idxmax(), "tag"
    ]

    url = "https://www.investing.com/crypto/" + tag

    head = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    req = requests.get(url, headers=head)

    if req.status_code != 200:
        raise ConnectionError(
            "ERR#0015: error " + str(req.status_code) + ", try again later."
        )

    root_ = fromstring(req.text)
    path_ = root_.xpath("//div[@class='cryptoGlobalData']/div")

    result = pd.DataFrame(
        columns=[
            "Crypto Currency",
            "Market Cap",
            "Circulating Supply",
            "Max Supply",
            "Vol (24H)",
            "Todays Range",
            "Chg (7D)",
            "Currency",
        ]
    )
    result.at[0, "Crypto Currency"] = name
    result.at[0, "Currency"] = currency

    if path_:
        for elements_ in path_:
            element = elements_.xpath(".//span[@class='title']")[0]
            title_ = element.text_content().replace(":", "")
            if title_ == "Day's Range":
                title_ = "Todays Range"
            if title_ in result.columns.tolist():
                result.at[0, title_] = element.getnext().text_content().strip()

        result.replace({"N/A": None}, inplace=True)

        if as_json is True:
            json_ = result.iloc[0].to_dict()
            return json_
        elif as_json is False:
            return result
    else:
        raise RuntimeError("ERR#0004: data retrieval error while scraping.")


def get_cryptos_overview(as_json=False, n_results=100):
    """
    This function retrieves an overview containing all the real time data available for the main crypto currencies,
    such as the names, symbols, current value, etc. as indexed in Investing.com. So on, the main usage of this
    function is to get an overview on the main crypto currencies, so to get a general view. Note that since
    this function is retrieving a lot of information at once, by default just the overview of the Top 100 crypto
    currencies is being retrieved, but an additional parameter called n_results can be specified so to retrieve N results.

    Args:
        as_json (:obj:`bool`, optional):
            optional argument to determine the format of the output data (:obj:`pandas.DataFrame` or :obj:`json`).
        n_results (:obj:`int`, optional):
            number of results to be displayed on the overview table (0-all_cryptos), where all crypto currencies will
            be retrieved if n_results=None.

    Note:
        The amount of indexed crypto currencies may vary, so if n_results is set to `None`, all the available crypto
        currencies in Investing.com while retrieving the overview, will be retrieved and returned.

    Returns:
        :obj:`pandas.DataFrame` - cryptos_overview:
            The resulting :obj:`pandas.DataFrame` contains all the data available in Investing.com of the main crypto
            currencies in order to get an overview of it.

            If the retrieval process succeeded, the resulting :obj:`pandas.DataFrame` should look like::

                name | symbol | price | market_cap | volume24h | total_volume | change24h | change7d | currency
                -----|--------|-------|------------|-----------|--------------|-----------|----------|----------
                xxxx | xxxxxx | xxxxx | xxxxxxxxxx | xxxxxxxxx | xxxxxxxxxxxx | xxxxxxxxx | xxxxxxxx | xxxxxxxx

    Raises:
        ValueError: raised if any of the introduced arguments is not valid or errored.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError: raised it no overview results could be retrieved from Investing.com.
        ConnectionError: raised if GET requests does not return 200 status code.

    """

    if not isinstance(as_json, bool):
        raise ValueError(
            "ERR#0002: as_json argument can just be True or False, bool type."
        )

    if n_results is not None and not isinstance(n_results, int):
        raise ValueError(
            "ERR#0089: n_results argument should be an integer between 1 and 1000."
        )

    if n_results is not None:
        if 1 > n_results or n_results > 1000:
            raise ValueError(
                "ERR#0089: n_results argument should be an integer between 1 and 1000."
            )

    header = {
        "User-Agent": random_user_agent(),
        "X-Requested-With": "XMLHttpRequest",
        "Accept": "text/html",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
    }

    url = "https://www.investing.com/crypto/currencies"

    req = requests.get(url, headers=header)

    root = fromstring(req.text)
    table = root.xpath(".//table[contains(@class, 'allCryptoTlb')]/tbody/tr")

    results = list()

    flag = False

    if len(table) < 1:
        raise RuntimeError(
            "ERR#0092: no data found while retrieving the overview from Investing.com"
        )

    if n_results is not None and n_results <= 100:
        table = table[:n_results]
        flag = True

    for row in table:
        name = (
            row.xpath(".//td[contains(@class, 'cryptoName')]")[0].text_content().strip()
        )
        symbol = row.xpath(".//td[contains(@class, 'symb')]")[0].get("title").strip()

        # Due to Investing.com parsing error
        if symbol in ["GRV", "GLYPH"]:
            continue

        tag = row.xpath(".//td[contains(@class, 'cryptoName')]/a")

        if len(tag) > 0:
            status = "available"
        else:
            status = "unavailable"

        price = row.xpath(".//td[contains(@class, 'price')]")[0].text_content()

        market_cap = row.xpath(".//td[@class='js-market-cap']")[0].get("data-value")
        volume24h = row.xpath(".//td[@class='js-24h-volume']")[0].get("data-value")
        total_volume = row.xpath(".//td[@class='js-total-vol']")[0].text_content()

        change24h = row.xpath(".//td[contains(@class, 'js-currency-change-24h')]")[
            0
        ].text_content()
        change7d = row.xpath(".//td[contains(@class, 'js-currency-change-7d')]")[
            0
        ].text_content()

        data = {
            "name": name,
            "symbol": symbol,
            "status": status,
            "price": float(price.replace(",", "")),
            "market_cap": float(market_cap.replace(",", "")),
            "volume24h": volume24h,
            "total_volume": total_volume,
            "change24h": change24h,
            "change7d": change7d,
            "currency": "USD",
        }

        results.append(data)

    if flag is True:
        df = pd.DataFrame(results)

        if as_json:
            return json.loads(df.to_json(orient="records"))
        else:
            return df
    else:
        header = {
            "User-Agent": random_user_agent(),
            "X-Requested-With": "XMLHttpRequest",
            "Accept": "text/html",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
        }

        params = {"lastRowId": 100}

        url = "https://www.investing.com/crypto/Service/LoadCryptoCurrencies"

        req = requests.post(url=url, headers=header, data=params)

        root = fromstring(req.json()["html"])
        table = root.xpath(".//tr")

        if n_results is not None:
            remaining_cryptos = n_results - len(results)
            table = table[:remaining_cryptos]

        if len(table) < 1:
            raise RuntimeError(
                "ERR#0092: no data found while retrieving the overview from"
                " Investing.com"
            )

        for row in table:
            name = (
                row.xpath(".//td[contains(@class, 'cryptoName')]")[0]
                .text_content()
                .strip()
            )
            symbol = (
                row.xpath(".//td[contains(@class, 'symb')]")[0].get("title").strip()
            )

            # Due to Investing.com parsing error
            if symbol in ["GRV", "GLYPH"]:
                continue

            tag = row.xpath(".//td[contains(@class, 'cryptoName')]/a")

            if len(tag) > 0:
                status = "available"
            else:
                status = "unavailable"

            price = row.xpath(".//td[contains(@class, 'price')]")[0].text_content()

            market_cap = row.xpath(".//td[@class='js-market-cap']")[0].get("data-value")
            volume24h = row.xpath(".//td[@class='js-24h-volume']")[0].get("data-value")
            total_volume = row.xpath(".//td[@class='js-total-vol']")[0].text_content()

            change24h = row.xpath(".//td[contains(@class, 'js-currency-change-24h')]")[
                0
            ].text_content()
            change7d = row.xpath(".//td[contains(@class, 'js-currency-change-7d')]")[
                0
            ].text_content()

            data = {
                "name": name,
                "symbol": symbol,
                "status": status,
                "price": float(price.replace(",", "")),
                "market_cap": float(market_cap.replace(",", "")),
                "volume24h": volume24h,
                "total_volume": total_volume,
                "change24h": change24h,
                "change7d": change7d,
                "currency": "USD",
            }

            results.append(data)

    df = pd.DataFrame(results)

    if as_json:
        return json.loads(df.to_json(orient="records"))
    else:
        return df


def search_cryptos(by, value):
    """
    This function searches cryptos by the introduced value for the specified field. This means that this function
    is going to search if there is a value that matches the introduced one for the specified field which is the
    `cryptos.csv` column name to search in. Available fields to search cryptos are 'name' and 'symbol'.

    Args:
        by (:obj:`str`): name of the field to search for, which is the column name which can be: 'name' or 'symbol'.
        value (:obj:`str`): value of the field to search for, which is the value that is going to be searched.

    Returns:
        :obj:`pandas.DataFrame` - search_result:
            The resulting :obj:`pandas.DataFrame` contains the search results from the given query, which is
            any match of the specified value in the specified field. If there are no results for the given query,
            an error will be raised, but otherwise the resulting :obj:`pandas.DataFrame` will contain all the
            available cryptos that match the introduced query.

    Raises:
        ValueError: raised if any of the introduced parameters is not valid or errored.
        FileNotFoundError: raised if `cryptos.csv` file is missing.
        IOError: raised if data could not be retrieved due to file error.
        RuntimeError: raised if no results were found for the introduced value in the introduced field.

    """

    if not by:
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not isinstance(by, str):
        raise ValueError(
            "ERR#0006: the introduced field to search is mandatory and should be a str."
        )

    if not value:
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    if not isinstance(value, str):
        raise ValueError(
            "ERR#0017: the introduced value to search is mandatory and should be a str."
        )

    resource_package = "investpy"
    resource_path = "/".join(("resources", "cryptos.csv"))
    if pkg_resources.resource_exists(resource_package, resource_path):
        cryptos = pd.read_csv(
            pkg_resources.resource_filename(resource_package, resource_path),
            keep_default_na=False,
        )
    else:
        raise FileNotFoundError("ERR#0081: cryptos file not found or errored.")

    if cryptos is None:
        raise IOError("ERR#0082: cryptos not found or unable to retrieve.")

    cryptos.drop(columns=["tag", "id"], inplace=True)

    available_search_fields = cryptos.columns.tolist()

    if isinstance(by, str) and by not in available_search_fields:
        raise ValueError(
            "ERR#0026: the introduced field to search can either just be "
            + " or ".join(available_search_fields)
        )

    cryptos["matches"] = cryptos[by].str.contains(value, case=False)

    search_result = cryptos.loc[cryptos["matches"] == True].copy()

    if len(search_result) == 0:
        raise RuntimeError(
            "ERR#0043: no results were found for the introduced " + str(by) + "."
        )

    search_result.drop(columns=["matches"], inplace=True)
    search_result.reset_index(drop=True, inplace=True)

    return search_result
